@@ -214,11 +214,14 @@ gif_read (GifContext *context, guchar *buffer, size_t len)
 #endif
 		retval = (fread(buffer, len, 1, context->file) != 0);
 
-                if (!retval && ferror (context->file))
+                if (!retval && ferror (context->file)) {
+                        gint save_errno = errno;
                         g_set_error (context->error,
                                      G_FILE_ERROR,
-                                     g_file_error_from_errno (errno),
-                                     _("Failure reading GIF: %s"), strerror (errno));
+                                     g_file_error_from_errno (save_errno),
+                                     _("Failure reading GIF: %s"), 
+                                     strerror (save_errno));
+                }
                 
 #ifdef IO_GIFDEBUG
 		if (len < 100) {
@@ -905,6 +908,11 @@ gif_get_lzw (GifContext *context)
 
                 /* Only call prepare_func for the first frame */
 		if (context->animation->frames->next == NULL) { 
+                        if (context->animation->width == 0 )
+                                context->animation->width = gdk_pixbuf_get_width(context->frame->pixbuf);
+                        if (context->animation->height == 0)
+                                context->animation->height = gdk_pixbuf_get_height (context->frame->pixbuf);
+
                         if (context->prepare_func)
                                 (* context->prepare_func) (context->frame->pixbuf,
                                                            GDK_PIXBUF_ANIMATION (context->animation),
@@ -1518,7 +1526,7 @@ gdk_pixbuf__gif_image_stop_load (gpointer data, GError **error)
 	GifContext *context = (GifContext *) data;
         gboolean retval = TRUE;
         
-        if (context->state != GIF_DONE) {
+        if (context->state != GIF_DONE || context->animation->frames == NULL) {
                 g_set_error (error,
                              GDK_PIXBUF_ERROR,
                              GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
@@ -1582,7 +1590,7 @@ gdk_pixbuf__gif_image_load_increment (gpointer data,
 		/* prepare for the next image_load_increment */
 		if (context->buf == buf) {
 			g_assert (context->size == size);
-			context->buf = (guchar *)g_new (guchar, context->amount_needed + (context->size - context->ptr));
+			context->buf = g_new (guchar, context->amount_needed + (context->size - context->ptr));
 			memcpy (context->buf, buf + context->ptr, context->size - context->ptr);
 		} else {
 			/* copy the left overs to the begining of the buffer */
@@ -1646,6 +1654,12 @@ gdk_pixbuf__gif_image_load_animation (FILE *file,
 	return animation;
 }
 
+#ifndef INCLUDE_gif
+#define MODULE_ENTRY(type,function) function
+#else
+#define MODULE_ENTRY(type,function) _gdk_pixbuf__ ## type ## _ ## function
+#endif
+
 void
 MODULE_ENTRY (gif, fill_vtable) (GdkPixbufModule *module)
 {
