@@ -228,8 +228,7 @@ static IOBuffer *io_buffer_free_segment(IOBuffer *buffer,
 static void io_buffer_free(IOBuffer *buffer)
 {
 	g_return_if_fail(buffer != NULL);
-	if (buffer->data)
-		g_free(buffer->data);
+	g_free(buffer->data);
 	g_free(buffer);
 }
 
@@ -504,8 +503,9 @@ static guint parse_rle_data_pseudocolor(TGAContext *ctx)
 		}
 	}
 
-	if (ctx->pbuf_bytes_done == ctx->pbuf_bytes)
+	if (ctx->pbuf_bytes_done == ctx->pbuf_bytes) 
 		ctx->done = TRUE;
+	
 	return n;
 }
 
@@ -632,18 +632,14 @@ static gboolean parse_rle_data(TGAContext *ctx, GError **err)
 {
 	guint rows = 0;
 	guint count = 0;
-	guint pbuf_count = 0;
 	guint bytes_done_before = ctx->pbuf_bytes_done;
-	if (ctx->hdr->type == TGA_TYPE_RLE_PSEUDOCOLOR) {
+
+	if (ctx->hdr->type == TGA_TYPE_RLE_PSEUDOCOLOR)
 		count = parse_rle_data_pseudocolor(ctx);
-		pbuf_count = count * ctx->pbuf->n_channels;
-	} else if (ctx->hdr->type == TGA_TYPE_RLE_TRUECOLOR) {
+	else if (ctx->hdr->type == TGA_TYPE_RLE_TRUECOLOR)
 		count = parse_rle_data_truecolor(ctx);
-		pbuf_count = count;
-	} else if (ctx->hdr->type == TGA_TYPE_RLE_GRAYSCALE) {
+	else if (ctx->hdr->type == TGA_TYPE_RLE_GRAYSCALE)
 		count = parse_rle_data_grayscale(ctx);
-		pbuf_count = count * (ctx->pbuf->n_channels == 4 ? 2 : 3);
-	}
 
 	if (ctx->hdr->flags & TGA_ORIGIN_RIGHT) {
 		guchar *row = ctx->pbuf->pixels + (bytes_done_before / ctx->pbuf->rowstride) * ctx->pbuf->rowstride;
@@ -661,8 +657,10 @@ static gboolean parse_rle_data(TGAContext *ctx, GError **err)
 		 * perfect, but doing it during the rle decoding in place
 		 * is considerably more work. 
 		 */
-		if (!(ctx->hdr->flags & TGA_ORIGIN_UPPER))
+		if (!(ctx->hdr->flags & TGA_ORIGIN_UPPER)) {
 			pixbuf_flip_vertically (ctx->pbuf);
+			ctx->hdr->flags |= TGA_ORIGIN_UPPER;
+		}
 
 	}
 		
@@ -937,11 +935,16 @@ static gboolean gdk_pixbuf__tga_stop_load(gpointer data, GError **err)
 	TGAContext *ctx = (TGAContext *) data;
 	g_return_val_if_fail(ctx != NULL, FALSE);
 
-	if (ctx->hdr)
-	  g_free (ctx->hdr);
+	if (!(ctx->hdr->flags & TGA_ORIGIN_UPPER) && ctx->run_length_encoded) {
+		pixbuf_flip_vertically (ctx->pbuf);
+		if (ctx->ufunc)
+			(*ctx->ufunc) (ctx->pbuf, 0, 0,
+				       ctx->pbuf->width, ctx->pbuf->height,
+			       	       ctx->udata);
+	}
+	g_free (ctx->hdr);
 	if (ctx->cmap) {
-	  if (ctx->cmap->cols)
-	    g_free (ctx->cmap->cols);
+	  g_free (ctx->cmap->cols);
 	  g_free (ctx->cmap);
 	}
 	if (ctx->pbuf)
@@ -957,6 +960,12 @@ static gboolean gdk_pixbuf__tga_stop_load(gpointer data, GError **err)
 	return TRUE;
 }
 
+#ifndef INCLUDE_tga
+#define MODULE_ENTRY(type,function) function
+#else
+#define MODULE_ENTRY(type,function) _gdk_pixbuf__ ## type ## _ ## function
+#endif
+
 void
 MODULE_ENTRY (tga, fill_vtable) (GdkPixbufModule *module)
 {
