@@ -37,12 +37,14 @@
 
 #include "gdkdirectfb.h"
 #include "gdkprivate-directfb.h"
+#include "gdkdisplay-directfb.h"
 
 #include "gdkregion-generic.h"
 
 #include "gdkinternals.h"
 #include "gdkalias.h"
 #include "cairo.h"
+#include <assert.h>
 
 static GdkRegion * gdk_window_impl_directfb_get_visible_region (GdkDrawable *drawable);
 static void        gdk_window_impl_directfb_set_colormap       (GdkDrawable *drawable,
@@ -59,11 +61,73 @@ typedef struct
 } GdkWindowChildHandlerData;
 
 
+/* Code for dirty-region queueing
+ */
+static GSList *update_windows = NULL;
+static guint update_idle = 0;
+
+static void
+gdk_window_directfb_process_all_updates (void)
+{
+  GSList *old_update_windows = update_windows;
+  GSList *tmp_list = update_windows;
+
+  if (update_idle)
+    g_source_remove (update_idle);
+  
+  update_windows = NULL;
+  update_idle = 0;
+
+  g_slist_foreach (old_update_windows, (GFunc)g_object_ref, NULL);
+  
+  while (tmp_list)
+    {
+      GdkWindowObject *private = (GdkWindowObject *)tmp_list->data;
+      
+      if (private->update_freeze_count)
+	update_windows = g_slist_prepend (update_windows, private);
+      else
+	gdk_window_process_updates(tmp_list->data,TRUE);
+      
+      g_object_unref (tmp_list->data);
+      tmp_list = tmp_list->next;
+    }
+
+  g_slist_free (old_update_windows);
+
+}
+
+static gboolean
+gdk_window_update_idle (gpointer data)
+{
+  gdk_window_directfb_process_all_updates ();
+  
+  return FALSE;
+}
+
+static void
+gdk_window_schedule_update (GdkWindow *window)
+{
+  if (window && GDK_WINDOW_OBJECT (window)->update_freeze_count)
+    return;
+
+  if (!update_idle)
+    {
+      update_idle = gdk_threads_add_idle_full (GDK_PRIORITY_REDRAW,
+				     gdk_window_update_idle, NULL, NULL);
+    }
+}
+
 
 static GdkWindow *gdk_directfb_window_containing_pointer = NULL;
 static GdkWindow *gdk_directfb_focused_window            = NULL;
 static gpointer   parent_class                           = NULL;
 GdkWindow * _gdk_parent_root = NULL;
+static void
+gdk_window_impl_directfb_paintable_init (GdkPaintableIface *iface);
+
+
+
 
 
 GType
@@ -86,9 +150,20 @@ gdk_window_impl_directfb_get_type (void)
           (GInstanceInitFunc) gdk_window_impl_directfb_init,
         };
 
+    static const GInterfaceInfo paintable_info =
+      {
+    (GInterfaceInitFunc) gdk_window_impl_directfb_paintable_init,
+    NULL,
+    NULL
+      };
+
       object_type = g_type_register_static (GDK_TYPE_DRAWABLE_IMPL_DIRECTFB,
                                             "GdkWindowImplDirectFB",
                                             &object_info, 0);
+       g_type_add_interface_static (object_type,
+                   GDK_TYPE_PAINTABLE,
+                   &paintable_info);
+
     }
 
   return object_type;
@@ -160,7 +235,6 @@ gdk_window_impl_directfb_finalize (GObject *object)
     {
       gdk_directfb_window_id_table_remove (impl->dfb_id);
 	  /* native window resource must be release before we can finalize !*/
-      /*impl->window->Release (impl->window);*/
       impl->window = NULL;
     }
 
@@ -176,7 +250,7 @@ gdk_window_impl_directfb_get_visible_region (GdkDrawable *drawable)
   DFBRectangle             drect = { 0, 0, 0, 0 };
 
   if (priv->surface)
-    priv->surface->GetVisibleRectangle (priv->surface, &drect);
+  priv->surface->GetVisibleRectangle (priv->surface, &drect);
   rect.x= drect.x;
   rect.y= drect.y;
   rect.width=drect.w;
@@ -218,12 +292,14 @@ create_directfb_window (GdkWindowImplDirectFB *impl,
   if (ret != DFB_OK)
     {
       DirectFBError ("gdk_window_new: Layer->CreateWindow failed", ret);
+      g_assert (0);
       return FALSE;
     }
 
   if ((desc->flags & DWDESC_CAPS) && (desc->caps & DWCAPS_INPUTONLY))
+  {
     impl->drawable.surface = NULL;
-  else
+  } else 
     window->GetSurface (window, &impl->drawable.surface);
 
   if (window_options)
@@ -257,7 +333,8 @@ _gdk_windowing_window_init (void)
   private->window_type = GDK_WINDOW_ROOT;
   private->state       = 0;
   private->children    = NULL;
-
+  impl->drawable.paint_region   = NULL;
+  impl->gdkWindow      = _gdk_parent_root;
   impl->window           = NULL;
   impl->drawable.abs_x   = 0;
   impl->drawable.abs_y   = 0;
@@ -269,10 +346,11 @@ _gdk_windowing_window_init (void)
     DFBWindowDescription   desc;
     desc.flags = 0;
 	/*XXX I must do this now its a bug  ALPHA ROOT*/
-#if 1 
-    desc.flags |= DWDESC_CAPS;
-    desc.caps = DWCAPS_ALPHACHANNEL;
-#endif
+
+    desc.flags = DWDESC_CAPS;
+    desc.caps = 0;
+    desc.caps  |= DWCAPS_NODECORATION;
+    desc.caps  |= DWCAPS_ALPHACHANNEL;
     desc.flags |= ( DWDESC_WIDTH | DWDESC_HEIGHT |
                       DWDESC_POSX  | DWDESC_POSY );
     desc.posx   = 0;
@@ -285,6 +363,10 @@ _gdk_windowing_window_init (void)
   }
   impl->drawable.surface->GetPixelFormat(impl->drawable.surface,&impl->drawable.format); 
   private->depth = DFB_BITS_PER_PIXEL(impl->drawable.format);
+  /*
+	Now we can set up the system colormap
+  */
+  gdk_drawable_set_colormap (GDK_DRAWABLE (_gdk_parent_root),gdk_colormap_get_system());
 }
 
 
@@ -325,6 +407,7 @@ gdk_directfb_window_new (GdkWindow              *parent,
 
   impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
   impl->drawable.wrapper = GDK_DRAWABLE (window);
+  impl->gdkWindow      = window;
 
   private->x = x;
   private->y = y;
@@ -414,9 +497,9 @@ gdk_directfb_window_new (GdkWindow              *parent,
 	   impl->window=NULL;
       if (!private->input_only && parent_impl->drawable.surface)
         {
+
           DFBRectangle rect =
           { x, y, impl->drawable.width, impl->drawable.height };
-
           parent_impl->drawable.surface->GetSubSurface (parent_impl->drawable.surface,
                                                         &rect,
                                                         &impl->drawable.surface);
@@ -476,6 +559,9 @@ gdk_directfb_window_new (GdkWindow              *parent,
       gdk_directfb_event_windows_add (window);
     }
 
+  if (attributes_mask & GDK_WA_TYPE_HINT)
+    gdk_window_set_type_hint (window, attributes->type_hint);
+
   return window;
 }
 
@@ -514,7 +600,9 @@ _gdk_windowing_window_destroy (GdkWindow *window,
   impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
 
   _gdk_selection_window_destroyed (window);
-
+#if (DIRECTFB_MAJOR_VERSION >= 1)
+  gdk_directfb_event_windows_remove (window);
+#endif
   if (window == _gdk_directfb_pointer_grab_window)
     gdk_pointer_ungrab (GDK_CURRENT_TIME);
   if (window == _gdk_directfb_keyboard_grab_window)
@@ -523,23 +611,23 @@ _gdk_windowing_window_destroy (GdkWindow *window,
   if (window == gdk_directfb_focused_window)
     gdk_directfb_change_focus (NULL);
 
-/*
-printf( " DESTROYING WINDOW %p %p %d parent=%p\n",window,impl->window,recursing,(GDK_WINDOW_OBJECT (window))->parent);
-*/
+
+  if (impl->drawable.surface) {
+    GdkDrawableImplDirectFB *dimpl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
+    if(dimpl->cairo_surface) {
+      cairo_surface_destroy(dimpl->cairo_surface);
+      dimpl->cairo_surface= NULL;
+    }
+    impl->drawable.surface->Release (impl->drawable.surface);
+    impl->drawable.surface = NULL;
+  }
+
   if (!recursing && !foreign_destroy && impl->window ) {
     	impl->window->SetOpacity (impl->window,0);
    		impl->window->Close(impl->window);
       	impl->window->Release(impl->window);
+        impl->window = NULL;
   }
-
-#if 0 /* let the finalizer kill it */
-  if (!recursing && !foreign_destroy)
-    {
-  		if (impl->window)
-    		impl->window->Destroy (impl->window);
-  		impl->window = NULL;
-	}
-#endif
 }
 
 /* This function is called when the window is really gone.
@@ -1183,11 +1271,15 @@ _gdk_directfb_move_resize_child (GdkWindow *window,
 
   if (!private->input_only)
     {
-      if (impl->drawable.surface)
-        {
-          impl->drawable.surface->Release (impl->drawable.surface);
-          impl->drawable.surface = NULL;
-        }
+    if (impl->drawable.surface) {
+      GdkDrawableImplDirectFB *dimpl = GDK_DRAWABLE_IMPL_DIRECTFB (private->impl);
+      if(dimpl->cairo_surface) {
+        cairo_surface_destroy(dimpl->cairo_surface);
+        dimpl->cairo_surface= NULL;
+      }
+    impl->drawable.surface->Release (impl->drawable.surface);
+    impl->drawable.surface = NULL;
+  }
 
       parent_impl = GDK_WINDOW_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (private->parent)->impl);
 
@@ -1291,6 +1383,9 @@ gdk_window_reparent (GdkWindow *window,
   GdkWindowObject *window_private;
   GdkWindowObject *parent_private;
   GdkWindowObject *old_parent_private;
+  GdkWindowImplDirectFB *impl;
+  GdkWindowImplDirectFB *parent_impl;
+  GdkVisual             *visual;
 
   g_return_if_fail (GDK_IS_WINDOW (window));
 
@@ -1303,6 +1398,12 @@ gdk_window_reparent (GdkWindow *window,
   window_private = (GdkWindowObject *) window;
   old_parent_private = (GdkWindowObject *) window_private->parent;
   parent_private = (GdkWindowObject *) new_parent;
+  parent_impl = GDK_WINDOW_IMPL_DIRECTFB (parent_private->impl);
+  visual = gdk_drawable_get_visual (window);
+
+  /* already parented */
+  if( window_private->parent == (GdkWindowObject *)new_parent )
+          return;
 
   window_private->parent = (GdkWindowObject *) new_parent;
 
@@ -1312,9 +1413,66 @@ gdk_window_reparent (GdkWindow *window,
         g_list_remove (old_parent_private->children, window);
     }
 
-  parent_private->children = g_list_prepend (parent_private->children, window);
+    parent_private->children = g_list_prepend (parent_private->children, window);
 
-  gdk_window_move_resize (window, x, y, -1, -1);
+    impl = GDK_WINDOW_IMPL_DIRECTFB (window_private->impl);
+
+    if( impl->drawable.surface ) {
+        impl->drawable.surface->Release (impl->drawable.surface);
+        impl->drawable.surface = NULL;
+    }
+
+    if( impl->window != NULL ) { 
+        gdk_directfb_window_id_table_remove (impl->dfb_id);
+    	impl->window->SetOpacity (impl->window,0);
+   		impl->window->Close(impl->window);
+      	impl->window->Release(impl->window);
+        impl->window = NULL;
+    }
+
+    //create window were a child of the root now
+    if( window_private->parent == (GdkWindowObject *)_gdk_parent_root)  {
+        DFBWindowDescription  desc;
+        DFBWindowOptions  window_options = DWOP_NONE;
+        desc.flags = DWDESC_CAPS;
+        if( window_private->input_only ) {
+            desc.caps = DWCAPS_INPUTONLY;
+        } else {
+            desc.flags |= DWDESC_PIXELFORMAT;
+            desc.pixelformat = ((GdkVisualDirectFB *) visual)->format;
+            if (DFB_PIXELFORMAT_HAS_ALPHA (desc.pixelformat)) {
+                desc.flags |= DWDESC_CAPS;
+                desc.caps = DWCAPS_ALPHACHANNEL;
+            }
+       }
+       if( window_private->window_type ==  GDK_WINDOW_CHILD )
+           window_private->window_type = GDK_WINDOW_TOPLEVEL;
+        desc.flags |= ( DWDESC_WIDTH | DWDESC_HEIGHT |
+                      DWDESC_POSX  | DWDESC_POSY );
+        desc.posx   = x;
+        desc.posy   = y;
+        desc.width  = impl->drawable.width;
+        desc.height = impl->drawable.height;
+        if (!create_directfb_window (impl, &desc, window_options))
+        {
+		  g_assert(0);
+          _gdk_window_destroy (window, FALSE);
+          return;
+        }
+        /* we hold a reference count on ourselves */
+        g_object_ref (window);
+        impl->window->GetID (impl->window, &impl->dfb_id);
+        gdk_directfb_window_id_table_insert (impl->dfb_id, window);
+        gdk_directfb_event_windows_add (window);
+   } else {
+          DFBRectangle rect = { x, y, impl->drawable.width,
+                                     impl->drawable.height};
+          impl->window = NULL;
+          parent_impl->drawable.surface->GetSubSurface (
+                          parent_impl->drawable.surface,
+                          &rect,
+                          &impl->drawable.surface);
+   }
 }
 
 void
@@ -1548,6 +1706,23 @@ gdk_window_set_role (GdkWindow   *window,
   /* N/A */
 }
 
+/**
+ * gdk_window_set_startup_id:
+ * @window: a toplevel #GdkWindow
+ * @startup_id: a string with startup-notification identifier
+ *
+ * When using GTK+, typically you should use gtk_window_set_startup_id()
+ * instead of this low-level function.
+ *
+ * Since: 2.12
+ *
+ **/
+void          
+gdk_window_set_startup_id (GdkWindow   *window,
+		     const gchar *startup_id)
+{
+}
+
 void
 gdk_window_set_transient_for (GdkWindow *window,
                               GdkWindow *parent)
@@ -2273,73 +2448,6 @@ gdk_window_set_static_gravities (GdkWindow *window,
 
   return FALSE;
 }
-#if 0 
-void
-gdk_window_begin_paint_region (GdkWindow *window,
-                               GdkRegion *region)
-{
-  GdkDrawableImplDirectFB *impl;
-  gint                     i;
-
-  g_return_if_fail (GDK_IS_WINDOW (window));
-
-  impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);
-
-  impl->buffered = TRUE;
-  impl->paint_depth++;
-
-  if (!region)
-    return;
-
-  if (impl->paint_region)
-    gdk_region_union (impl->paint_region, region);
-  else
-    impl->paint_region = gdk_region_copy (region);
-
-  for (i = 0; i < region->numRects; i++)
-    {
-      GdkRegionBox *box = &region->rects[i];
-
-      _gdk_windowing_window_clear_area (window,
-                                        box->x1,
-                                        box->y1,
-                                        box->x2 - box->x1,
-                                        box->y2 - box->y1);
-    }
-}
-
-void
-gdk_window_end_paint (GdkWindow *window)
-{
-  GdkDrawableImplDirectFB *impl;
-
-  g_return_if_fail (GDK_IS_WINDOW (window));
-
-  impl = GDK_DRAWABLE_IMPL_DIRECTFB (GDK_WINDOW_OBJECT (window)->impl);
-
-  g_return_if_fail (impl->paint_depth > 0);
-
-  impl->paint_depth--;
-
-  if (impl->paint_depth == 0)
-    {
-      impl->buffered = FALSE;
-
-      if (impl->paint_region)
-        {
-          DFBRegion reg = { impl->paint_region->extents.x1,
-                            impl->paint_region->extents.y1,
-                            impl->paint_region->extents.x2 - 1,
-                            impl->paint_region->extents.y2 - 1 };
-
-          _gdk_directfb_update (impl, &reg);
-
-          gdk_region_destroy (impl->paint_region);
-          impl->paint_region = NULL;
-        }
-    }
-}
-#endif
 
 void
 gdk_window_begin_resize_drag (GdkWindow     *window,
@@ -2411,19 +2519,157 @@ gdk_window_get_frame_extents (GdkWindow    *window,
   rect->height = impl->height;
 }
 
+/*
+ * Given a directfb window and a subsurface of that window
+ * create a gdkwindow child wrapper
+ */
+#if (DIRECTFB_MAJOR_VERSION >= 1)
+GdkWindow *gdk_directfb_create_child_window(GdkWindow *parent,
+                                IDirectFBSurface *subsurface)
+{
+  GdkWindow             *window;
+  GdkWindowObject       *private;
+  GdkWindowObject       *parent_private;
+  GdkWindowImplDirectFB *impl;
+  GdkWindowImplDirectFB *parent_impl;
+  gint x,y,w,h;
 
+  g_return_val_if_fail (parent != NULL, NULL);
 
+  window = g_object_new (GDK_TYPE_WINDOW, NULL);
+  private = GDK_WINDOW_OBJECT (window);
+  parent_private = GDK_WINDOW_OBJECT (parent);
+  parent_impl = GDK_WINDOW_IMPL_DIRECTFB (parent_private->impl);
+  private->parent = parent_private;
+
+  subsurface->GetPosition(subsurface,&x,&y);
+  subsurface->GetSize(subsurface,&w,&h);
+
+  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
+  impl->drawable.wrapper = GDK_DRAWABLE (window);
+
+  private->x = x;
+  private->y = y;
+
+  _gdk_directfb_calc_abs (window);
+
+  impl->drawable.width  = w;
+  impl->drawable.height = h;
+  private->window_type = GDK_WINDOW_CHILD;
+  impl->drawable.surface = subsurface;
+  impl->drawable.format = parent_impl->drawable.format;
+  private->depth = parent_private->depth;
+  gdk_drawable_set_colormap (GDK_DRAWABLE (window),
+        gdk_drawable_get_colormap (parent));
+  gdk_window_set_cursor (window, NULL);
+  parent_private->children = g_list_prepend (parent_private->children,window);
+  /*we hold a reference count on ourselves */
+  g_object_ref (window);
+
+  return window;
+
+}
+#endif
+
+/*
+ * The wrapping is not perfect since directfb does not give full access
+ * to the current state of a window event mask etc need to fix dfb
+ */
 GdkWindow *
 gdk_window_foreign_new_for_display (GdkDisplay* display,GdkNativeWindow anid)
 {
-  GdkWindow *window;
+    GdkWindow *window = NULL;
+    GdkWindow              *parent =NULL;
+    GdkWindowObject       *private =NULL;
+    GdkWindowObject       *parent_private =NULL;
+    GdkWindowImplDirectFB *parent_impl =NULL;
+    GdkWindowImplDirectFB *impl =NULL;
+    DFBWindowOptions options;
+    DFBResult        ret;
+    GdkDisplayDFB * gdkdisplay =  _gdk_display;
+    IDirectFBWindow *dfbwindow;
 
-  window = gdk_window_lookup (anid);
+    window = gdk_window_lookup (anid);
 
-  if (window)
+    if (window) {
+        g_object_ref (window);
+        return window;
+    }
+    if( display != NULL )
+            gdkdisplay = GDK_DISPLAY_DFB(display);
+
+    ret = gdkdisplay->layer->GetWindow (gdkdisplay->layer,
+                    (DFBWindowID)anid,&dfbwindow);
+
+    if (ret != DFB_OK) {
+        DirectFBError ("gdk_window_new: Layer->GetWindow failed", ret);
+        return NULL;
+    }
+
+    parent = _gdk_parent_root;
+
+    if(parent) {
+        parent_private = GDK_WINDOW_OBJECT (parent);
+        parent_impl = GDK_WINDOW_IMPL_DIRECTFB (parent_private->impl);
+    }
+
+    window = g_object_new (GDK_TYPE_WINDOW, NULL);
+    /* we hold a reference count on ourselves */
     g_object_ref (window);
+    private = GDK_WINDOW_OBJECT (window);
+    private->parent = parent_private;
+    private->window_type = GDK_WINDOW_TOPLEVEL;
+    impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
 
-  return window;
+    impl->drawable.wrapper = GDK_DRAWABLE (window);
+    impl->window = dfbwindow;
+    dfbwindow->GetOptions(dfbwindow,&options);
+    dfbwindow->GetPosition(dfbwindow,&private->x,&private->y);
+    dfbwindow->GetSize(dfbwindow,&impl->drawable.width,&impl->drawable.height);
+
+
+    private->input_only = FALSE;
+
+    if( dfbwindow->GetSurface (dfbwindow, &impl->drawable.surface) == DFB_UNSUPPORTED ){
+        private->input_only = TRUE;
+        impl->drawable.surface = NULL;
+    }
+    /*
+     * Position ourselevs
+     */
+    _gdk_directfb_calc_abs (window);
+
+    /* We default to all events least surprise to the user 
+     * minus the poll for motion events 
+     */
+    gdk_window_set_events (window, (GDK_ALL_EVENTS_MASK ^ GDK_POINTER_MOTION_HINT_MASK));
+
+    if (impl->drawable.surface)
+    {
+      impl->drawable.surface->GetPixelFormat (impl->drawable.surface,
+					      &impl->drawable.format);
+
+  	  private->depth = DFB_BITS_PER_PIXEL(impl->drawable.format);
+      if( parent )
+        gdk_drawable_set_colormap (GDK_DRAWABLE (window), gdk_drawable_get_colormap (parent));
+      else
+        gdk_drawable_set_colormap (GDK_DRAWABLE (window), gdk_colormap_get_system());
+    }
+
+    //can  be null for the soft cursor window itself when
+    //running a gtk directfb wm
+    if( gdk_display_get_default() != NULL ) {
+        gdk_window_set_cursor (window,NULL);
+    }
+
+    if (parent_private)
+        parent_private->children = g_list_prepend (parent_private->children,
+                                               window);
+    impl->dfb_id = (DFBWindowID)anid; 
+    gdk_directfb_window_id_table_insert (impl->dfb_id, window);
+    gdk_directfb_event_windows_add (window);
+
+    return window;
 }
 
 GdkWindow *
@@ -2438,13 +2684,30 @@ gdk_window_lookup (GdkNativeWindow anid)
   return gdk_directfb_window_id_table_lookup ((DFBWindowID) anid);
 }
 
+IDirectFBWindow *gdk_directfb_window_lookup(GdkWindow *window )
+{
+  GdkWindowObject       *private;
+  GdkWindowImplDirectFB *impl;
+  g_return_val_if_fail (GDK_IS_WINDOW (window),NULL);
+  private = GDK_WINDOW_OBJECT (window);
+  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
+  return impl->window;
+}
 
+IDirectFBSurface *gdk_directfb_surface_lookup(GdkWindow *window)
+{
+  GdkWindowObject       *private;
+  GdkWindowImplDirectFB *impl;
+  g_return_val_if_fail (GDK_IS_WINDOW (window),NULL);
+  private = GDK_WINDOW_OBJECT (window);
+  impl = GDK_WINDOW_IMPL_DIRECTFB (private->impl);
+  return impl->drawable.surface;
+}
 
 void
 gdk_window_fullscreen (GdkWindow *window)
 {
   g_return_if_fail (GDK_IS_WINDOW (window));
-
   g_warning ("gdk_window_fullscreen() not implemented.\n");
 }
 
@@ -2452,8 +2715,7 @@ void
 gdk_window_unfullscreen (GdkWindow *window)
 {
   g_return_if_fail (GDK_IS_WINDOW (window));
-
-  /*  g_warning ("gdk_window_unfullscreen() not implemented.\n");*/
+  /* g_warning ("gdk_window_unfullscreen() not implemented.\n");*/
 }
 
 void
@@ -2513,6 +2775,278 @@ gdk_window_set_urgency_hint (GdkWindow *window,
 
 }
 
+static void
+gdk_window_impl_directfb_invalidate_maybe_recurse (GdkPaintable *paintable,
+                         GdkRegion    *region,
+                         gboolean    (*child_func) (GdkWindow *, gpointer),
+                         gpointer      user_data)
+{
+  GdkWindow *window;
+  GdkWindowObject *private;
+  GdkWindowImplDirectFB *wimpl;
+  GdkDrawableImplDirectFB *impl;
+
+  wimpl = GDK_WINDOW_IMPL_DIRECTFB (paintable);
+  impl = (GdkDrawableImplDirectFB *)wimpl;
+  window = wimpl->gdkWindow;
+  private = (GdkWindowObject *)window;
+
+  GdkRegion *visible_region;
+  GList *tmp_list;
+
+  g_return_if_fail (window != NULL);
+  g_return_if_fail (GDK_IS_WINDOW (window));
+
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+  
+  if (private->input_only || !GDK_WINDOW_IS_MAPPED (window))
+    return;
+
+  visible_region = gdk_drawable_get_visible_region (window);
+  gdk_region_intersect (visible_region, region);
+
+  tmp_list = private->children;
+  while (tmp_list)
+    {
+      GdkWindowObject *child = tmp_list->data;
+      
+      if (!child->input_only)
+	{
+	  GdkRegion *child_region;
+	  GdkRectangle child_rect;
+	  
+	  gdk_window_get_position ((GdkWindow *)child,
+				   &child_rect.x, &child_rect.y);
+	  gdk_drawable_get_size ((GdkDrawable *)child,
+				 &child_rect.width, &child_rect.height);
+
+	  child_region = gdk_region_rectangle (&child_rect);
+	  
+	  /* remove child area from the invalid area of the parent */
+	  if (GDK_WINDOW_IS_MAPPED (child) && !child->shaped)
+	    gdk_region_subtract (visible_region, child_region);
+	  
+	  if (child_func && (*child_func) ((GdkWindow *)child, user_data))
+	    {
+	      gdk_region_offset (region, - child_rect.x, - child_rect.y);
+	      gdk_region_offset (child_region, - child_rect.x, - child_rect.y);
+	      gdk_region_intersect (child_region, region);
+	      
+	      gdk_window_invalidate_maybe_recurse ((GdkWindow *)child,
+						   child_region, child_func, user_data);
+	      
+	      gdk_region_offset (region, child_rect.x, child_rect.y);
+	    }
+
+	  gdk_region_destroy (child_region);
+	}
+
+      tmp_list = tmp_list->next;
+    }
+  
+  if (!gdk_region_empty (visible_region))
+    {
+      
+      if (private->update_area)
+	{
+	  gdk_region_union (private->update_area, visible_region);
+	}
+      else
+	{
+	  update_windows = g_slist_prepend (update_windows, window);
+	  private->update_area = gdk_region_copy (visible_region);
+	  gdk_window_schedule_update (window);
+	}
+    }
+    gdk_region_destroy (visible_region);
+}
+
+
+static void
+gdk_window_impl_directfb_process_updates (GdkPaintable *paintable,
+                    gboolean      update_children)
+{
+  GdkWindow *window;
+  GdkWindowObject *private;
+  GdkWindowImplDirectFB *wimpl;
+  GdkDrawableImplDirectFB *impl;
+
+  wimpl = GDK_WINDOW_IMPL_DIRECTFB (paintable);
+  impl = (GdkDrawableImplDirectFB *)wimpl;
+  window = wimpl->gdkWindow;
+  private = (GdkWindowObject *)window;
+  gboolean save_region = FALSE;
+
+  /* If an update got queued during update processing, we can get a
+   * window in the update queue that has an empty update_area.
+   * just ignore it.
+   */
+  if (private->update_area)
+    {
+      GdkRegion *update_area = private->update_area;
+      private->update_area = NULL;
+      
+      if (_gdk_event_func && gdk_window_is_viewable (window))
+	{
+	  GdkRectangle window_rect;
+	  GdkRegion *expose_region;
+	  GdkRegion *window_region;
+          gint width, height;
+	  save_region = _gdk_windowing_window_queue_antiexpose (window, update_area);
+
+	  if (save_region)
+	    expose_region = gdk_region_copy (update_area);
+	  else
+	    expose_region = update_area;
+	  
+          gdk_drawable_get_size (GDK_DRAWABLE (private), &width, &height);
+
+	  window_rect.x = 0;
+	  window_rect.y = 0;
+	  window_rect.width = width;
+	  window_rect.height = height;
+
+	  window_region = gdk_region_rectangle (&window_rect);
+	  gdk_region_intersect (expose_region,
+				window_region);
+	  gdk_region_destroy (window_region);
+	  
+	  if (!gdk_region_empty (expose_region) &&
+	      (private->event_mask & GDK_EXPOSURE_MASK))
+	    {
+	      GdkEvent event;
+	      
+	      event.expose.type = GDK_EXPOSE;
+	      event.expose.window = g_object_ref (window);
+	      event.expose.send_event = FALSE;
+	      event.expose.count = 0;
+	      event.expose.region = expose_region;
+	      gdk_region_get_clipbox (expose_region, &event.expose.area);
+	      (*_gdk_event_func) (&event, _gdk_event_data);
+	      
+	      g_object_unref (window);
+	    }
+
+	  if (expose_region != update_area)
+	    gdk_region_destroy (expose_region);
+	}
+      if (!save_region)
+	gdk_region_destroy (update_area);
+    }
+}
+
+
+static void
+gdk_window_impl_directfb_begin_paint_region (GdkPaintable *paintable,
+					   GdkRegion    *region)
+{
+  GdkDrawableImplDirectFB *impl;
+  GdkWindowImplDirectFB *wimpl;
+  gint                     i;
+
+
+  g_assert (region != NULL );
+  wimpl = GDK_WINDOW_IMPL_DIRECTFB (paintable);
+  impl = (GdkDrawableImplDirectFB *)wimpl;
+  impl->buffered = TRUE;
+  impl->paint_depth++;
+
+  if (!region)
+    return;
+
+  if (impl->paint_region)
+    gdk_region_union (impl->paint_region, region);
+  else
+    impl->paint_region = gdk_region_copy (region);
+
+  for (i = 0; i < region->numRects; i++)
+    {
+      GdkRegionBox *box = &region->rects[i];
+
+      _gdk_windowing_window_clear_area (GDK_WINDOW(wimpl->gdkWindow),
+                                        box->x1,
+                                        box->y1,
+                                        box->x2 - box->x1,
+                                        box->y2 - box->y1);
+                                        
+    }
+}
+
+static void
+gdk_window_impl_directfb_end_paint (GdkPaintable *paintable)
+{
+  GdkDrawableImplDirectFB *impl;
+
+  impl = GDK_DRAWABLE_IMPL_DIRECTFB (paintable);
+
+  g_return_if_fail (impl->paint_depth > 0);
+
+  impl->paint_depth--;
+
+  if (impl->paint_depth == 0)
+    {
+      impl->buffered = FALSE;
+
+      if (impl->paint_region)
+        {
+          DFBRegion reg = { impl->paint_region->extents.x1,
+                            impl->paint_region->extents.y1,
+                            impl->paint_region->extents.x2 ,
+                            impl->paint_region->extents.y2 };
+
+          impl->surface->Flip(impl->surface, &reg,0);
+          gdk_region_destroy (impl->paint_region);
+          impl->paint_region = NULL;
+        }
+    }
+}
+
+
+static void
+gdk_window_impl_directfb_paintable_init (GdkPaintableIface *iface)
+{
+  iface->begin_paint_region = gdk_window_impl_directfb_begin_paint_region;
+  iface->end_paint = gdk_window_impl_directfb_end_paint;
+
+  iface->invalidate_maybe_recurse = gdk_window_impl_directfb_invalidate_maybe_recurse;
+  iface->process_updates = gdk_window_impl_directfb_process_updates;
+}
+
+void
+gdk_window_beep (GdkWindow *window)
+{
+  gdk_display_beep (gdk_display_get_default());
+}
+
+void
+gdk_window_set_opacity (GdkWindow *window,
+			gdouble    opacity)
+{
+  GdkDisplay *display;
+  guint8 cardinal;
+  
+  g_return_if_fail (GDK_IS_WINDOW (window));
+
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
+  display = gdk_drawable_get_display (window);
+
+  if (opacity < 0)
+    opacity = 0;
+  else if (opacity > 1)
+    opacity = 1;
+  cardinal = opacity * 0xff;
+  gdk_directfb_window_set_opacity(window,cardinal);
+}
+
+void
+_gdk_windowing_window_set_composited (GdkWindow *window,
+                                      gboolean   composited)
+{
+}
+
 
 #define __GDK_WINDOW_X11_C__
 #include "gdkaliasdef.c"
