@@ -1,6 +1,7 @@
 /* GDK - The GIMP Drawing Kit
  * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
  * Copyright (C) 1998-2002 Tor Lillqvist
+ * Copyright (C) 2007 Cody Russell
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -534,6 +535,8 @@ gdk_pointer_grab (GdkWindow    *window,
   
   if (return_val == GDK_GRAB_SUCCESS)
     {
+      GdkWindowImplWin32 *impl = GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl);
+
       if (p_grab_window != NULL && p_grab_window != window)
 	generate_grab_broken_event (p_grab_window, FALSE, window);
       
@@ -550,6 +553,10 @@ gdk_pointer_grab (GdkWindow    *window,
 
       if (p_grab_cursor != NULL)
 	SetCursor (p_grab_cursor);
+      else if (impl->hcursor != NULL)
+	SetCursor (impl->hcursor);
+      else
+	SetCursor (LoadCursor (NULL, IDC_ARROW));
 
       if (confine_to != NULL)
 	{
@@ -2810,7 +2817,17 @@ gdk_event_translate (MSG  *msg,
       event->any.window = window;
 
       append_event (event);
-      
+
+      if (event->any.type == GDK_UNMAP)
+	{
+	  impl = GDK_WINDOW_IMPL_WIN32 (GDK_WINDOW_OBJECT (window)->impl);
+
+	  if (impl->transient_owner && GetForegroundWindow () == GDK_WINDOW_HWND (window))
+	    {
+	      SetForegroundWindow (GDK_WINDOW_HWND (impl->transient_owner));
+	    }
+	}
+
       if (event->any.type == GDK_UNMAP &&
 	  p_grab_window == window)
 	gdk_pointer_ungrab (msg->time);
@@ -3207,6 +3224,13 @@ gdk_event_translate (MSG  *msg,
 
       append_event (event);
 
+      impl = GDK_WINDOW_IMPL_WIN32 (GDK_WINDOW_OBJECT (window)->impl);
+
+      if (impl->transient_owner && GetForegroundWindow() == GDK_WINDOW_HWND (window))
+	{
+	  SetForegroundWindow (GDK_WINDOW_HWND (impl->transient_owner));
+	}
+
       return_val = TRUE;
       break;
 
