@@ -40,15 +40,14 @@
 
 guint _gdk_keymap_serial = 0;
 gboolean _gdk_keyboard_has_altgr = FALSE;
+guint _scancode_rshift = 0;
+
 static GdkModifierType gdk_shift_modifiers = GDK_SHIFT_MASK;
 
 static GdkKeymap *default_keymap = NULL;
 
 static guint *keysym_tab = NULL;
 
-typedef int (WINAPI *t_ToUnicodeEx) (UINT,UINT,PBYTE,LPWSTR,int,UINT,HKL);
-static t_ToUnicodeEx p_ToUnicodeEx = NULL;
-
 #ifdef G_ENABLE_DEBUG
 static void
 print_keysym_tab (void)
@@ -142,6 +141,8 @@ handle_special (guint  vk,
       *ksymp = GDK_Meta_L; break;
     case VK_RWIN:
       *ksymp = GDK_Meta_R; break;
+    case VK_APPS:
+      *ksymp = GDK_Menu; break;
     case VK_MULTIPLY:
       *ksymp = GDK_KP_Multiply; break;
     case VK_ADD:
@@ -152,6 +153,26 @@ handle_special (guint  vk,
       *ksymp = GDK_KP_Subtract; break;
     case VK_DIVIDE:
       *ksymp = GDK_KP_Divide; break;
+    case VK_NUMPAD0:
+      *ksymp = GDK_KP_0; break;
+    case VK_NUMPAD1:
+      *ksymp = GDK_KP_1; break;
+    case VK_NUMPAD2:
+      *ksymp = GDK_KP_2; break;
+    case VK_NUMPAD3:
+      *ksymp = GDK_KP_3; break;
+    case VK_NUMPAD4:
+      *ksymp = GDK_KP_4; break;
+    case VK_NUMPAD5:
+      *ksymp = GDK_KP_5; break;
+    case VK_NUMPAD6:
+      *ksymp = GDK_KP_6; break;
+    case VK_NUMPAD7:
+      *ksymp = GDK_KP_7; break;
+    case VK_NUMPAD8:
+      *ksymp = GDK_KP_8; break;
+    case VK_NUMPAD9:
+      *ksymp = GDK_KP_9; break;
     case VK_F1:
       *ksymp = GDK_F1; break;
     case VK_F2:
@@ -242,6 +263,7 @@ static void
 reset_after_dead (guchar key_state[256])
 {
   guchar temp_key_state[256];
+  wchar_t wcs[2];
 
   memmove (temp_key_state, key_state, sizeof (key_state));
 
@@ -249,23 +271,12 @@ reset_after_dead (guchar key_state[256])
     temp_key_state[VK_CONTROL] =
     temp_key_state[VK_MENU] = 0;
 
-  if (G_WIN32_HAVE_WIDECHAR_API ())
-    {
-      wchar_t wcs[2];
-      (*p_ToUnicodeEx) (VK_SPACE, MapVirtualKey (VK_SPACE, 0),
-			temp_key_state, wcs, G_N_ELEMENTS (wcs),
-			0, _gdk_input_locale);
-    }
-  else
-    {
-      char chars[2];
-      ToAsciiEx (VK_SPACE, MapVirtualKey (VK_SPACE, 0),
-		 temp_key_state, (LPWORD) chars, 0,
-		 _gdk_input_locale);
-    }
+  ToUnicodeEx (VK_SPACE, MapVirtualKey (VK_SPACE, 0),
+	       temp_key_state, wcs, G_N_ELEMENTS (wcs),
+	       0, _gdk_input_locale);
 }
 
-static gboolean
+static void
 handle_dead (guint  keysym,
 	     guint *ksymp)
 {
@@ -306,9 +317,12 @@ handle_dead (guint  keysym,
     case GDK_Greek_accentdieresis: /* 0x7ae */
       *ksymp = GDK_Greek_accentdieresis; break;
     default:
-      return FALSE;
+      /* By default use the keysym as such. This takes care of for
+       * instance the dead U+09CD (BENGALI VIRAMA) on the ekushey
+       * Bengali layout.
+       */
+      *ksymp = keysym; break;
     }
-  return TRUE;
 }
 
 static void
@@ -319,22 +333,10 @@ update_keymap (void)
   guint scancode;
   guint vk;
   gboolean capslock_tested = FALSE;
-  static HMODULE user32 = NULL;
 
   if (keysym_tab != NULL && current_serial == _gdk_keymap_serial)
     return;
 
-  g_assert (G_WIN32_HAVE_WIDECHAR_API () || _gdk_input_codepage != 0);
-
-  if (G_WIN32_HAVE_WIDECHAR_API () && user32 == NULL)
-    {
-      user32 = GetModuleHandle ("user32.dll");
-
-      g_assert (user32 != NULL);
-
-      p_ToUnicodeEx = (t_ToUnicodeEx) GetProcAddress (user32, "ToUnicodeEx");
-    }
-
   current_serial = _gdk_keymap_serial;
 
   if (keysym_tab == NULL)
@@ -347,7 +349,8 @@ update_keymap (void)
 
   for (vk = 0; vk < 256; vk++)
     {
-      if ((scancode = MapVirtualKey (vk, 0)) == 0)
+      if ((scancode = MapVirtualKey (vk, 0)) == 0 &&
+	  vk != VK_DIVIDE)
 	keysym_tab[vk*4+0] =
 	  keysym_tab[vk*4+1] =
 	  keysym_tab[vk*4+2] =
@@ -356,6 +359,9 @@ update_keymap (void)
 	{
 	  gint shift;
 
+	  if (vk == VK_RSHIFT)
+	    _scancode_rshift = scancode;
+
 	  key_state[vk] = 0x80;
 	  for (shift = 0; shift < 4; shift++)
 	    {
@@ -378,50 +384,17 @@ update_keymap (void)
 		  wchar_t wcs[10];
 		  gint k;
 
-		  if (G_WIN32_HAVE_WIDECHAR_API ())
-		    {
-		      k = (*p_ToUnicodeEx) (vk, scancode, key_state,
-					    wcs, G_N_ELEMENTS (wcs),
-					    0, _gdk_input_locale);
+		  k = ToUnicodeEx (vk, scancode, key_state,
+				   wcs, G_N_ELEMENTS (wcs),
+				   0, _gdk_input_locale);
 #if 0
-		      g_print ("ToUnicodeEx(%02x, %d: %d): %d, %04x %04x\n",
-			       vk, scancode, shift, k,
-			       (k != 0 ? wcs[0] : 0),
-			       (k >= 2 ? wcs[1] : 0));
+		  g_print ("ToUnicodeEx(%02x, %d: %d): %d, %04x %04x\n",
+			   vk, scancode, shift, k,
+			   (k != 0 ? wcs[0] : 0),
+			   (k >= 2 ? wcs[1] : 0));
 #endif
-		      if (k == 1)
-			*ksymp = gdk_unicode_to_keyval (wcs[0]);
-		    }
-		  else
-		    {
-		      k = ToAsciiEx (vk, scancode, key_state,
-				     (LPWORD) chars, 0, _gdk_input_locale);
-#if 0
-		      g_print ("ToAsciiEx(%02x, %d: %d): %d, %02x %02x\n",
-			       vk, scancode, shift, k,
-			       (k != 0 ? chars[0] : 0),
-			       (k == 2 ? chars[1] : 0));
-#endif
-		      if (k == 1)
-			{
-			  if (_gdk_input_codepage >= 1250 &&
-			      _gdk_input_codepage <= 1258 &&
-			      chars[0] >= GDK_space &&
-			      chars[0] <= GDK_asciitilde)
-			    *ksymp = chars[0];
-			  else
-			    {
-			      if (MultiByteToWideChar (_gdk_input_codepage, 0,
-						       chars, 1, wcs, 1) > 0)
-				*ksymp = gdk_unicode_to_keyval (wcs[0]);
-			    }
-			}
-		      else if (k == -1)
-			{
-			  MultiByteToWideChar (_gdk_input_codepage, 0,
-					       chars, 1, wcs, 1);
-			}
-		    }
+		  if (k == 1)
+		    *ksymp = gdk_unicode_to_keyval (wcs[0]);
 
 		  if (k == 1)
 		    {
@@ -445,14 +418,7 @@ update_keymap (void)
 		      reset_after_dead (key_state);
 
 		      /* Use dead keysyms instead of "undead" ones */
-		      if (!handle_dead (keysym, ksymp))
-			GDK_NOTE (EVENTS,
-				  g_print ("Unhandled dead key cp:%d vk:%02x sc:%x ch:%02x wc:%04x keysym:%04x%s%s\n",
-					   _gdk_input_codepage, vk,
-					   scancode, chars[0],
-					   wcs[0], keysym,
-					   (shift&0x1 ? " shift" : ""),
-					   (shift&0x2 ? " altgr" : "")));
+		      handle_dead (keysym, ksymp);
 		    }
 		  else if (k == 0)
 		    {
@@ -465,10 +431,8 @@ update_keymap (void)
 		    {
 #if 0
 		      GDK_NOTE (EVENTS,
-				g_print ("%s returns %d "
+				g_print ("ToUnicodeEx returns %d "
 					 "for vk:%02x, sc:%02x%s%s\n",
-					 (G_WIN32_HAVE_WIDECHAR_API () ?
-					  "ToUnicodeEx" : "ToAsciiEx"),
 					 k, vk, scancode,
 					 (shift&0x1 ? " shift" : ""),
 					 (shift&0x2 ? " altgr" : "")));
@@ -562,6 +526,16 @@ gdk_keymap_get_direction (GdkKeymap *keymap)
     }
 }
 
+gboolean
+gdk_keymap_have_bidi_layouts (GdkKeymap *keymap)
+{
+  /* Should we check if the kayboard layouts switchable at the moment
+   * cover both directionalities? What does the doc comment in
+   * ../x11/gdkkeys-x11.c exactly mean?
+   */
+  return FALSE;
+}
+
 gboolean
 gdk_keymap_get_entries_for_keyval (GdkKeymap     *keymap,
                                    guint          keyval,
@@ -891,7 +865,7 @@ gdk_keymap_translate_keyboard_state (GdkKeymap       *keymap,
     }
 				
 #if 0
-  GDK_NOTE (EVENTS, g_print ("...group=%d level=%d cmods=%#x keyval=%s\n",
+  GDK_NOTE (EVENTS, g_print ("... group=%d level=%d cmods=%#x keyval=%s\n",
 			     group, shift_level, tmp_modifiers, gdk_keyval_name (tmp_keyval)));
 #endif
 
