@@ -24,6 +24,8 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
  */
 
+#include <config.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -38,8 +40,7 @@
 #include "gdkinternals.h"
 #include "gdkdisplay-x11.h"
 #include "gdkkeysyms.h"
-
-#include "config.h"
+#include "gdkalias.h"
 
 #ifdef HAVE_XKB
 #include <X11/XKBlib.h>
@@ -52,16 +53,26 @@
 #  endif
 #endif /* HAVE_XKB */
 
-typedef struct _GdkKeymapX11 GdkKeymapX11;
+typedef struct _GdkKeymapX11   GdkKeymapX11;
+typedef struct _GdkKeymapClass GdkKeymapX11Class;
 
 #define GDK_TYPE_KEYMAP_X11          (gdk_keymap_x11_get_type ())
 #define GDK_KEYMAP_X11(object)       (G_TYPE_CHECK_INSTANCE_CAST ((object), GDK_TYPE_KEYMAP_X11, GdkKeymapX11))
 #define GDK_IS_KEYMAP_X11(object)    (G_TYPE_CHECK_INSTANCE_TYPE ((object), GDK_TYPE_KEYMAP_X11))
 
+typedef struct _DirectionCacheEntry DirectionCacheEntry;
+
+struct _DirectionCacheEntry
+{
+  guint serial;
+  Atom group_atom;
+  PangoDirection direction;
+};
+
 struct _GdkKeymapX11
 {
   GdkKeymap     parent_instance;
-  
+
   gint min_keycode;
   gint max_keycode;
   KeySym* keymap;
@@ -70,21 +81,35 @@ struct _GdkKeymapX11
   guint lock_keysym;
   GdkModifierType group_switch_mask;
   GdkModifierType num_lock_mask;
+  GdkModifierType modmap[8];
   gboolean sun_keypad;
   PangoDirection current_direction;
   gboolean have_direction;
   guint current_serial;
-  
+
 #ifdef HAVE_XKB
   XkbDescPtr xkb_desc;
+  /* We cache the directions */
+  Atom current_group_atom;
+  guint current_cache_serial;
+  /* A cache of size four should be more than enough, people usually
+   * have two groups around, and the xkb limit is four.  It still
+   * works correct for more than four groups.  It's just the
+   * cache.
+   */
+  DirectionCacheEntry group_direction_cache[4];
 #endif
 };
 
 #define KEYMAP_USE_XKB(keymap) GDK_DISPLAY_X11 ((keymap)->display)->use_xkb
 #define KEYMAP_XDISPLAY(keymap) GDK_DISPLAY_XDISPLAY ((keymap)->display)
 
-static GType gdk_keymap_x11_get_type (void);
-static void  gdk_keymap_x11_init     (GdkKeymapX11 *keymap);
+static GType gdk_keymap_x11_get_type   (void);
+static void  gdk_keymap_x11_class_init (GdkKeymapX11Class *klass);
+static void  gdk_keymap_x11_init       (GdkKeymapX11      *keymap);
+static void  gdk_keymap_x11_finalize   (GObject           *object);
+
+static GdkKeymapClass *parent_class = NULL;
 
 static GType
 gdk_keymap_x11_get_type (void)
@@ -98,7 +123,7 @@ gdk_keymap_x11_get_type (void)
 	  sizeof (GdkKeymapClass),
 	  (GBaseInitFunc) NULL,
 	  (GBaseFinalizeFunc) NULL,
-	  (GClassInitFunc) NULL,
+	  (GClassInitFunc) gdk_keymap_x11_class_init,
 	  NULL,           /* class_finalize */
 	  NULL,           /* class_data */
 	  sizeof (GdkKeymapX11),
@@ -107,13 +132,23 @@ gdk_keymap_x11_get_type (void)
 	};
       
       object_type = g_type_register_static (GDK_TYPE_KEYMAP,
-                                            "GdkKeymapX11",
+                                            g_intern_static_string ("GdkKeymapX11"),
                                             &object_info, 0);
     }
   
   return object_type;
 }
 
+static void
+gdk_keymap_x11_class_init (GdkKeymapX11Class *klass)
+{
+  GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+  parent_class = g_type_class_peek_parent (klass);
+
+  object_class->finalize = gdk_keymap_x11_finalize;
+}
+
 static void
 gdk_keymap_x11_init (GdkKeymapX11 *keymap)
 {
@@ -129,12 +164,33 @@ gdk_keymap_x11_init (GdkKeymapX11 *keymap)
   keymap->group_switch_mask = 0;
   keymap->lock_keysym = GDK_Caps_Lock;
   keymap->have_direction = FALSE;
-  
+  keymap->current_serial = 0;
+
 #ifdef HAVE_XKB
   keymap->xkb_desc = NULL;
+  keymap->current_group_atom = 0;
+  keymap->current_cache_serial = 0;
 #endif
 
-  keymap->current_serial = 0;
+}
+
+static void
+gdk_keymap_x11_finalize (GObject *object)
+{
+  GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (object);
+
+  if (keymap_x11->keymap)
+    XFree (keymap_x11->keymap);
+
+  if (keymap_x11->mod_keymap)
+    XFreeModifiermap (keymap_x11->mod_keymap);
+
+#ifdef HAVE_XKB
+  if (keymap_x11->xkb_desc)
+    XkbFreeKeyboard (keymap_x11->xkb_desc, XkbAllComponentsMask, True);
+#endif
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static inline void
@@ -147,6 +203,46 @@ update_keyrange (GdkKeymapX11 *keymap_x11)
 
 #ifdef HAVE_XKB
 
+static void
+update_modmap (Display      *display,
+	       GdkKeymapX11 *keymap_x11)
+{
+  static struct {
+    const gchar *name;
+    Atom atom;
+    GdkModifierType mask;
+  } vmods[] = {
+    { "Meta", 0, GDK_META_MASK },
+    { "Super", 0, GDK_SUPER_MASK },
+    { "Hyper", 0, GDK_HYPER_MASK },
+    { NULL, 0, 0 }
+  };
+
+  gint i, j, k;
+
+  if (!vmods[0].atom)
+    for (i = 0; vmods[i].name; i++)
+      vmods[i].atom = XInternAtom (display, vmods[i].name, FALSE);
+
+  for (i = 0; i < 8; i++)
+    keymap_x11->modmap[i] = 1 << i;
+
+  for (i = 0; i < XkbNumVirtualMods; i++)
+    {
+      for (j = 0; vmods[j].atom; j++)
+	{
+	  if (keymap_x11->xkb_desc->names->vmods[i] == vmods[j].atom)
+	    {
+	      for (k = 0; k < 8; k++)
+		{
+		  if (keymap_x11->xkb_desc->server->vmods[i] & (1 << k))
+		    keymap_x11->modmap[k] |= vmods[j].mask;
+		}
+	    }
+	}
+    }
+}
+
 static XkbDescPtr
 get_xkb (GdkKeymapX11 *keymap_x11)
 {
@@ -157,17 +253,24 @@ get_xkb (GdkKeymapX11 *keymap_x11)
   
   if (keymap_x11->xkb_desc == NULL)
     {
-      keymap_x11->xkb_desc = XkbGetMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask, XkbUseCoreKbd);
+      keymap_x11->xkb_desc = XkbGetMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask | XkbModifierMapMask | XkbVirtualModsMask, XkbUseCoreKbd);
       if (keymap_x11->xkb_desc == NULL)
-	g_error ("Failed to get keymap");
+        {
+	  g_error ("Failed to get keymap");
+          return NULL;
+        }
 
-      XkbGetNames (xdisplay, XkbGroupNamesMask, keymap_x11->xkb_desc);
+      XkbGetNames (xdisplay, XkbGroupNamesMask | XkbVirtualModNamesMask, keymap_x11->xkb_desc);
+
+      update_modmap (xdisplay, keymap_x11);
     }
   else if (keymap_x11->current_serial != display_x11->keymap_serial)
     {
-      XkbGetUpdatedMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask,
+      XkbGetUpdatedMap (xdisplay, XkbKeySymsMask | XkbKeyTypesMask | XkbModifierMapMask | XkbVirtualModsMask,
 			keymap_x11->xkb_desc);
-      XkbGetNames (xdisplay, XkbGroupNamesMask, keymap_x11->xkb_desc);
+      XkbGetNames (xdisplay, XkbGroupNamesMask | XkbVirtualModNamesMask, keymap_x11->xkb_desc);
+
+      update_modmap (xdisplay, keymap_x11);
     }
 
   keymap_x11->current_serial = display_x11->keymap_serial;
@@ -206,12 +309,45 @@ gdk_keymap_get_for_display (GdkDisplay *display)
 }
 
 /* Find the index of the group/level pair within the keysyms for a key.
+ * We round up the number of keysyms per keycode to the next even number,
+ * otherwise we lose a whole group of keys
  */
 #define KEYSYM_INDEX(keymap_impl, group, level) \
-  (2 * ((group) % (keymap_impl->keysyms_per_keycode / 2)) + (level))
+  (2 * ((group) % (gint)((keymap_impl->keysyms_per_keycode + 1) / 2)) + (level))
 #define KEYSYM_IS_KEYPAD(s) (((s) >= 0xff80 && (s) <= 0xffbd) || \
                              ((s) >= 0x11000000 && (s) <= 0x1100ffff))
 
+static gint
+get_symbol (const KeySym *syms,
+	    GdkKeymapX11 *keymap_x11,
+	    gint group,
+	    gint level)
+{
+  gint index;
+
+  index = KEYSYM_INDEX(keymap_x11, group, level);
+  if (index >= keymap_x11->keysyms_per_keycode)
+      return NoSymbol;
+
+  return syms[index];
+}
+
+static void
+set_symbol (KeySym       *syms, 
+	    GdkKeymapX11 *keymap_x11, 
+	    gint          group, 
+	    gint          level, 
+	    KeySym        sym)
+{
+  gint index;
+
+  index = KEYSYM_INDEX(keymap_x11, group, level);
+  if (index >= keymap_x11->keysyms_per_keycode)
+      return;
+
+  syms[index] = sym;
+}
+
 static void
 update_keymaps (GdkKeymapX11 *keymap_x11)
 {
@@ -256,24 +392,24 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
 	  /* Check both groups */
 	  for (i = 0 ; i < 2 ; i++)
 	    {
-	      if (syms[KEYSYM_INDEX (keymap_x11, i, 0)] == GDK_Tab)
-		syms[KEYSYM_INDEX (keymap_x11, i, 1)] = GDK_ISO_Left_Tab;
+	      if (get_symbol (syms, keymap_x11, i, 0) == GDK_Tab)
+		set_symbol (syms, keymap_x11, i, 1, GDK_ISO_Left_Tab);
 	    }
 
           /*
            * If there is one keysym and the key symbol has upper and lower
            * case variants fudge the keymap
            */
-          if (syms[KEYSYM_INDEX (keymap_x11, 0, 1)] == 0)
+          if (get_symbol (syms, keymap_x11, 0, 1) == 0)
             {
               guint lower;
               guint upper;
 
-              gdk_keyval_convert_case (syms[KEYSYM_INDEX (keymap_x11, 0, 0)], &lower, &upper);
+              gdk_keyval_convert_case (get_symbol (syms, keymap_x11, 0, 0), &lower, &upper);
               if (lower != upper)
                 {
-                  syms[KEYSYM_INDEX (keymap_x11, 0, 0)] = lower;
-                  syms[KEYSYM_INDEX (keymap_x11, 0, 1)] = upper;
+		  set_symbol (syms, keymap_x11, 0, 0, lower);
+		  set_symbol (syms, keymap_x11, 0, 1, upper);
                 }
             }
       
@@ -287,6 +423,9 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
       keymap_x11->group_switch_mask = 0;
       keymap_x11->num_lock_mask = 0;
 
+      for (i = 0; i < 8; i++)
+	keymap_x11->modmap[i] = 1 << i;
+      
       /* There are 8 sets of modifiers, with each set containing
        * max_keypermod keycodes.
        */
@@ -306,6 +445,22 @@ update_keymaps (GdkKeymapX11 *keymap_x11)
 
           syms = keymap_x11->keymap + (keycode - keymap_x11->min_keycode) * keymap_x11->keysyms_per_keycode;
 
+	  mask = 0;
+	  for (j = 0; j < keymap_x11->keysyms_per_keycode; j++)
+	    {
+	      if (syms[j] == GDK_Meta_L ||
+		  syms[j] == GDK_Meta_R)
+		mask |= GDK_META_MASK;
+	      else if (syms[j] == GDK_Hyper_L ||
+		       syms[j] == GDK_Hyper_R)
+		mask |= GDK_HYPER_MASK;
+	      else if (syms[j] == GDK_Super_L ||
+		       syms[j] == GDK_Super_R)
+		mask |= GDK_SUPER_MASK;
+	    }
+
+	  keymap_x11->modmap[i/keymap_x11->mod_keymap->max_keypermod] |= mask;
+
           /* The fourth modifier, GDK_MOD1_MASK is 1 << 3.
 	   * Each group of max_keypermod entries refers to the same modifier.
            */
@@ -378,53 +533,170 @@ get_keymap (GdkKeymapX11 *keymap_x11)
   return keymap_x11->keymap;
 }
 
-#if HAVE_XKB
-static PangoDirection
-get_direction (GdkKeymapX11 *keymap_x11)
+#define GET_EFFECTIVE_KEYMAP(keymap) get_effective_keymap ((keymap), G_STRFUNC)
+
+static GdkKeymap *
+get_effective_keymap (GdkKeymap  *keymap,
+		      const char *function)
 {
-  XkbDescRec *xkb = get_xkb (keymap_x11);
-  const char *name;
-  XkbStateRec state_rec;
-  PangoDirection result;
-
-  GdkDisplay *display = GDK_KEYMAP (keymap_x11)->display;
-
-  XkbGetState (GDK_DISPLAY_XDISPLAY (display), XkbUseCoreKbd, &state_rec);
-  
-  if (xkb->names->groups[state_rec.locked_group] == None)
-    result = PANGO_DIRECTION_LTR;
-  else
+  if (!keymap)
     {
-      name = gdk_x11_get_xatom_name_for_display (display, xkb->names->groups[state_rec.locked_group]);
-
-      if (g_ascii_strcasecmp (name, "arabic") == 0 ||
-	  g_ascii_strcasecmp (name, "hebrew") == 0 ||
-	  g_ascii_strcasecmp (name, "israelian") == 0)
-	result = PANGO_DIRECTION_RTL;
-      else
-	result = PANGO_DIRECTION_LTR;
+      GDK_NOTE (MULTIHEAD,
+		g_message ("reverting to default display keymap in %s",
+			   function));
+      return gdk_keymap_get_default ();
     }
-    
-  return result;
+
+  return keymap;
 }
 
+#if HAVE_XKB
+static PangoDirection
+get_direction (XkbDescRec *xkb, 
+	       gint group)
+{
+  gint code;
+
+  gint rtl_minus_ltr = 0; /* total number of RTL keysyms minus LTR ones */
+
+  for (code = xkb->min_key_code; code <= xkb->max_key_code; code++)
+    {
+      gint level = 0;
+      KeySym sym = XkbKeySymEntry (xkb, code, level, group);
+      PangoDirection dir = pango_unichar_direction (gdk_keyval_to_unicode (sym));
+
+      switch (dir)
+	{
+	case PANGO_DIRECTION_RTL:
+	  rtl_minus_ltr++;
+	  break;
+	case PANGO_DIRECTION_LTR:
+	  rtl_minus_ltr--;
+	  break;
+	default:
+	  break;
+	}
+    }
+
+  if (rtl_minus_ltr > 0)
+    return PANGO_DIRECTION_RTL;
+  else
+    return PANGO_DIRECTION_LTR;
+}
+
+static PangoDirection
+get_direction_from_cache (GdkKeymapX11 *keymap_x11,
+			  XkbDescPtr xkb,
+			  gint group)
+{
+  Atom group_atom = xkb->names->groups[group];
+
+  gboolean cache_hit = FALSE;
+  DirectionCacheEntry *cache = keymap_x11->group_direction_cache;
+
+  PangoDirection direction = PANGO_DIRECTION_NEUTRAL;
+  gint i;
+
+  if (keymap_x11->have_direction)
+    {
+      /* lookup in cache */
+      for (i = 0; i < G_N_ELEMENTS (keymap_x11->group_direction_cache); i++)
+      {
+	if (cache[i].group_atom == group_atom)
+	  {
+	    cache_hit = TRUE;
+	    cache[i].serial = keymap_x11->current_cache_serial++; /* freshen */
+	    direction = cache[i].direction;
+	    group_atom = cache[i].group_atom;
+	    break;
+	  }
+      }
+    }
+  else
+    {
+      /* initialize cache */
+      for (i = 0; i < G_N_ELEMENTS (keymap_x11->group_direction_cache); i++)
+	{
+	  cache[i].group_atom = 0;
+	  cache[i].serial = keymap_x11->current_cache_serial;
+	}
+      keymap_x11->current_cache_serial++;
+    }
+
+  /* insert in cache */
+  if (!cache_hit)
+    {
+      gint oldest = 0;
+
+      direction = get_direction (xkb, group);
+
+      /* remove the oldest entry */
+      for (i = 0; i < G_N_ELEMENTS (keymap_x11->group_direction_cache); i++)
+	{
+	  if (cache[i].serial < cache[oldest].serial)
+	    oldest = i;
+	}
+      
+      cache[oldest].group_atom = group_atom;
+      cache[oldest].direction = direction;
+      cache[oldest].serial = keymap_x11->current_cache_serial++;
+    }
+
+  return direction;
+}
+
+static int
+get_num_groups (GdkKeymap *keymap,
+		XkbDescPtr xkb)
+{
+      Display *display = KEYMAP_XDISPLAY (keymap);
+      XkbGetControls(display, XkbSlowKeysMask, xkb);
+      XkbGetUpdatedMap (display, XkbKeySymsMask | XkbKeyTypesMask |
+			XkbModifierMapMask | XkbVirtualModsMask, xkb);
+      return xkb->ctrls->num_groups;
+}
+
+static void
+update_direction (GdkKeymapX11 *keymap_x11,
+		  gint          group)
+{
+  XkbDescPtr xkb = get_xkb (keymap_x11);
+  Atom group_atom;
+
+  group_atom = xkb->names->groups[group];
+
+  /* a group change? */
+  if (!keymap_x11->have_direction || keymap_x11->current_group_atom != group_atom)
+    {
+      keymap_x11->current_direction = get_direction_from_cache (keymap_x11, xkb, group);
+      keymap_x11->current_group_atom = group_atom;
+      keymap_x11->have_direction = TRUE;
+    }
+}
+
+/* keep this in sync with the XkbSelectEventDetails() call 
+ * in gdk_display_open()
+ */
 void
-_gdk_keymap_state_changed (GdkDisplay *display)
+_gdk_keymap_state_changed (GdkDisplay *display,
+			   XEvent     *xevent)
 {
   GdkDisplayX11 *display_x11 = GDK_DISPLAY_X11 (display);
+  XkbEvent *xkb_event = (XkbEvent *)xevent;
   
   if (display_x11->keymap)
     {
       GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (display_x11->keymap);
+      gboolean had_direction;
+      PangoDirection direction;
       
-      PangoDirection new_direction = get_direction (keymap_x11);
+      had_direction = keymap_x11->have_direction;
+      direction = keymap_x11->current_direction;
       
-      if (!keymap_x11->have_direction || new_direction != keymap_x11->current_direction)
-	{
-	  keymap_x11->have_direction = TRUE;
-	  keymap_x11->current_direction = new_direction;
-	  g_signal_emit_by_name (keymap_x11, "direction_changed");
-	}
+      update_direction (keymap_x11, XkbStateGroup (&xkb_event->state));
+      
+      if (!had_direction || direction != keymap_x11->current_direction)
+	g_signal_emit_by_name (keymap_x11, "direction_changed");      
     }
 }
 
@@ -441,33 +713,82 @@ _gdk_keymap_keys_changed (GdkDisplay *display)
     g_signal_emit_by_name (display_x11->keymap, "keys_changed", 0);
 }
 
+/** 
+ * gdk_keymap_get_direction:
+ * @keymap: a #GdkKeymap or %NULL to use the default keymap
+ *
+ * Returns the direction of effective layout of the keymap.
+ *
+ * @Returns: %PANGO_DIRECTION_LTR or %PANGO_DIRECTION_RTL if determines the
+ * direction.  %PANGO_DIRECTION_NEUTRAL otherwise.
+ **/
 PangoDirection
 gdk_keymap_get_direction (GdkKeymap *keymap)
 {
-  if (!keymap)
-    {
-      keymap = gdk_keymap_get_for_display (gdk_display_get_default ());
-      GDK_NOTE (MULTIHEAD,
-		g_message ("_multihead : reverting to default display keymap "
-			   "in gdk_keymap_get_direction"));
-    }
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
   
 #if HAVE_XKB
   if (KEYMAP_USE_XKB (keymap))
     {
       GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (keymap);
-      
+
       if (!keymap_x11->have_direction)
 	{
-	  keymap_x11->current_direction = get_direction (keymap_x11);
-	  keymap_x11->have_direction = TRUE;
+	  GdkDisplay *display = GDK_KEYMAP (keymap_x11)->display;
+	  XkbStateRec state_rec;
+
+	  XkbGetState (GDK_DISPLAY_XDISPLAY (display), XkbUseCoreKbd, 
+		       &state_rec);
+	  update_direction (keymap_x11, XkbStateGroup (&state_rec));
 	}
   
       return keymap_x11->current_direction;
     }
   else
 #endif /* HAVE_XKB */
-    return PANGO_DIRECTION_LTR;
+    return PANGO_DIRECTION_NEUTRAL;
+}
+
+/** 
+ * gdk_keymap_have_bidi_layouts:
+ * @keymap: a #GdkKeymap or %NULL to use the default keymap
+ *
+ * Determines if keyboard layouts for both right-to-left and left-to-right
+ * languages are in use.
+ *
+ * @Returns: %TRUE if there are layouts in both directions, %FALSE otherwise
+ *
+ * Since: 2.12
+ **/
+gboolean
+gdk_keymap_have_bidi_layouts (GdkKeymap *keymap)
+{
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
+
+#if HAVE_XKB
+  if (KEYMAP_USE_XKB (keymap))
+    {
+      GdkKeymapX11 *keymap_x11 = GDK_KEYMAP_X11 (keymap);
+      XkbDescPtr xkb = get_xkb (keymap_x11);
+      int num_groups = get_num_groups (keymap, xkb);
+
+      int i;
+      gboolean have_ltr_keyboard = FALSE;
+      gboolean have_rtl_keyboard = FALSE;
+
+      for (i = 0; i < num_groups; i++)
+      {
+	if (get_direction_from_cache (keymap_x11, xkb, i) == PANGO_DIRECTION_RTL)
+	  have_rtl_keyboard = TRUE;
+	else
+	  have_ltr_keyboard = TRUE;
+      }
+
+      return have_ltr_keyboard && have_rtl_keyboard;
+    }
+  else
+#endif /* HAVE_XKB */
+    return FALSE;
 }
 
 /**
@@ -505,14 +826,7 @@ gdk_keymap_get_entries_for_keyval (GdkKeymap     *keymap,
   g_return_val_if_fail (n_keys != NULL, FALSE);
   g_return_val_if_fail (keyval != 0, FALSE);
 
-  if (!keymap)
-    {
-      keymap = gdk_keymap_get_for_display (gdk_display_get_default ());
-      GDK_NOTE (MULTIHEAD,
-		g_message ("_multihead : reverting to default display keymap "
-			   "in gdk_keymap_get_entries_for_keyval\n"));
-    }
-
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
   keymap_x11 = GDK_KEYMAP_X11 (keymap);
   
   retval = g_array_new (FALSE, FALSE, sizeof (GdkKeymapKey));
@@ -631,8 +945,8 @@ gdk_keymap_get_entries_for_keyval (GdkKeymap     *keymap,
  * gdk_keymap_get_entries_for_keycode:
  * @keymap: a #GdkKeymap or %NULL to use the default keymap
  * @hardware_keycode: a keycode
- * @keys: return location for array of #GdkKeymapKey, or NULL
- * @keyvals: return location for array of keyvals, or NULL
+ * @keys: return location for array of #GdkKeymapKey, or %NULL
+ * @keyvals: return location for array of keyvals, or %NULL
  * @n_entries: length of @keys and @keyvals
  *
  * Returns the keyvals bound to @hardware_keycode.
@@ -659,14 +973,7 @@ gdk_keymap_get_entries_for_keycode (GdkKeymap     *keymap,
   g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), FALSE);
   g_return_val_if_fail (n_entries != NULL, FALSE);
 
-  if (!keymap)
-    {
-      keymap = gdk_keymap_get_for_display (gdk_display_get_default ());
-      GDK_NOTE (MULTIHEAD,
-		g_message ("_multihead : reverting to default display keymap "
-			   "in gdk_keymap_get_entries_for_keycode\n"));
-    }
-
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
   keymap_x11 = GDK_KEYMAP_X11 (keymap);
 
   update_keyrange (keymap_x11);
@@ -775,36 +1082,20 @@ gdk_keymap_get_entries_for_keycode (GdkKeymap     *keymap,
           ++i;
         }
     }
+
+  *n_entries = 0;
+
+  if (keys)
+    {
+      *n_entries = key_array->len;
+      *keys = (GdkKeymapKey*) g_array_free (key_array, FALSE);
+    }
   
-  if ((key_array && key_array->len > 0) ||
-      (keyval_array && keyval_array->len > 0))
+  if (keyvals)
     {
-      if (keys)
-        *keys = (GdkKeymapKey*) key_array->data;
-
-      if (keyvals)
-        *keyvals = (guint*) keyval_array->data;
-
-      if (key_array)
-        *n_entries = key_array->len;
-      else
-        *n_entries = keyval_array->len;
+      *n_entries = keyval_array->len;
+      *keyvals = (guint*) g_array_free (keyval_array, FALSE);
     }
-  else
-    {
-      if (keys)
-        *keys = NULL;
-
-      if (keyvals)
-        *keyvals = NULL;
-      
-      *n_entries = 0;
-    }
-
-  if (key_array)
-    g_array_free (key_array, key_array->len > 0 ? FALSE : TRUE);
-  if (keyval_array)
-    g_array_free (keyval_array, keyval_array->len > 0 ? FALSE : TRUE);
 
   return *n_entries > 0;
 }
@@ -832,15 +1123,8 @@ gdk_keymap_lookup_key (GdkKeymap          *keymap,
   g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), 0);
   g_return_val_if_fail (key != NULL, 0);
   g_return_val_if_fail (key->group < 4, 0);
-  
-  if (!keymap)
-    {
-      keymap = gdk_keymap_get_for_display (gdk_display_get_default ());
-      GDK_NOTE (MULTIHEAD,
-		g_message ("_multihead : reverting to default display keymap "
-			   "in gdk_keymap_lookup_key\n"));
-    }
 
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
   keymap_x11 = GDK_KEYMAP_X11 (keymap);
   
 #ifdef HAVE_XKB
@@ -855,7 +1139,7 @@ gdk_keymap_lookup_key (GdkKeymap          *keymap,
     {
       const KeySym *map = get_keymap (keymap_x11);
       const KeySym *syms = map + (key->keycode - keymap_x11->min_keycode) * keymap_x11->keysyms_per_keycode;
-      return syms [KEYSYM_INDEX (keymap_x11, key->group, key->level)];
+      return get_symbol (syms, keymap_x11, key->group, key->level);
     }
 }
 
@@ -987,7 +1271,7 @@ MyEnhancedXkbTranslateKeyCode(register XkbDescPtr     xkb,
     
     /* ---- End stuff GDK adds to the original Xlib version ---- */
     
-    return (syms[col]!=NoSymbol);
+    return (syms[col] != NoSymbol);
 }
 #endif /* HAVE_XKB */
 
@@ -1005,7 +1289,7 @@ translate_keysym (GdkKeymapX11   *keymap_x11,
   const KeySym *map = get_keymap (keymap_x11);
   const KeySym *syms = map + (hardware_keycode - keymap_x11->min_keycode) * keymap_x11->keysyms_per_keycode;
 
-#define SYM(k,g,l) syms[KEYSYM_INDEX (k,g,l)]
+#define SYM(k,g,l) get_symbol (syms, k,g,l)
 
   GdkModifierType shift_modifiers;
   gint shift_level;
@@ -1054,7 +1338,7 @@ translate_keysym (GdkKeymapX11   *keymap_x11,
     }
   else
     {
-      /* Fall back to the the first level if no symbol for the level
+      /* Fall back to the first level if no symbol for the level
        * we were passed.
        */
       shift_level = (state & shift_modifiers) ? 1 : 0;
@@ -1088,11 +1372,11 @@ translate_keysym (GdkKeymapX11   *keymap_x11,
  * @hardware_keycode: a keycode
  * @state: a modifier state 
  * @group: active keyboard group
- * @keyval: return location for keyval
- * @effective_group: return location for effective group
- * @level: return location for level
- * @consumed_modifiers: return location for modifiers that were used to determine the group or level
- * 
+ * @keyval: return location for keyval, or %NULL
+ * @effective_group: return location for effective group, or %NULL
+ * @level: return location for level, or %NULL
+ * @consumed_modifiers: return location for modifiers that were used to
+ *     determine the group or level, or %NULL
  *
  * Translates the contents of a #GdkEventKey into a keyval, effective
  * group, and level. Modifiers that affected the translation and
@@ -1165,7 +1449,8 @@ gdk_keymap_translate_keyboard_state (GdkKeymap       *keymap,
 
   g_return_val_if_fail (keymap == NULL || GDK_IS_KEYMAP (keymap), FALSE);
   g_return_val_if_fail (group < 4, FALSE);
-  
+
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);  
   keymap_x11 = GDK_KEYMAP_X11 (keymap);
 
   if (keyval)
@@ -1182,7 +1467,7 @@ gdk_keymap_translate_keyboard_state (GdkKeymap       *keymap,
   if (hardware_keycode < keymap_x11->min_keycode ||
       hardware_keycode > keymap_x11->max_keycode)
     return FALSE;
-  
+
 #ifdef HAVE_XKB
   if (KEYMAP_USE_XKB (keymap))
     {
@@ -1329,3 +1614,68 @@ _gdk_x11_get_group_for_state (GdkDisplay      *display,
       return (state & keymap_impl->group_switch_mask) ? 1 : 0;
     }
 }
+
+void
+_gdk_keymap_add_virtual_modifiers (GdkKeymap       *keymap,
+				   GdkModifierType *modifiers)
+{
+  GdkKeymapX11 *keymap_x11;
+  int i;
+  
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);
+  keymap_x11 = GDK_KEYMAP_X11 (keymap);
+
+  for (i = 3; i < 8; i++)
+    {
+      if ((1 << i) & *modifiers)
+        {
+	  if (keymap_x11->modmap[i] & GDK_MOD1_MASK)
+	    *modifiers |= GDK_MOD1_MASK;
+	  else if (keymap_x11->modmap[i] & GDK_SUPER_MASK)
+	    *modifiers |= GDK_SUPER_MASK;
+	  else if (keymap_x11->modmap[i] & GDK_HYPER_MASK)
+	    *modifiers |= GDK_HYPER_MASK;
+	  else if (keymap_x11->modmap[i] & GDK_META_MASK)
+	    *modifiers |= GDK_META_MASK;
+        }
+    }
+}
+
+gboolean
+_gdk_keymap_key_is_modifier (GdkKeymap *keymap,
+			     guint      keycode)
+{
+  GdkKeymapX11 *keymap_x11;
+  gint i;
+
+  keymap = GET_EFFECTIVE_KEYMAP (keymap);  
+  keymap_x11 = GDK_KEYMAP_X11 (keymap);
+
+  if (keycode < keymap_x11->min_keycode ||
+      keycode > keymap_x11->max_keycode)
+    return FALSE;
+
+#ifdef HAVE_XKB
+  if (KEYMAP_USE_XKB (keymap))
+    {
+      XkbDescRec *xkb = get_xkb (keymap_x11);
+      
+      if (xkb->map->modmap[keycode] != 0)
+	return TRUE;
+    }
+  else
+#endif
+    {
+      for (i = 0; i < 8 * keymap_x11->mod_keymap->max_keypermod; i++)
+	{
+	  if (keycode == keymap_x11->mod_keymap->modifiermap[i])
+	    return TRUE;
+	}
+    }
+
+  return FALSE;
+}
+
+
+#define __GDK_KEYS_X11_C__
+#include "gdkaliasdef.c"
