@@ -30,6 +30,7 @@
 #include <stdlib.h>
 
 #include "gtkcontainer.h"
+#include "gtkbuildable.h"
 #include "gtkprivate.h"
 #include "gtkmain.h"
 #include "gtkmarshalers.h"
@@ -99,6 +100,24 @@ static void     gtk_container_unmap                (GtkWidget         *widget);
 static gchar* gtk_container_child_default_composite_name (GtkContainer *container,
 							  GtkWidget    *child);
 
+/* GtkBuildable */
+static void gtk_container_buildable_init           (GtkBuildableIface *iface);
+static void gtk_container_buildable_add_child      (GtkBuildable *buildable,
+						    GtkBuilder   *builder,
+						    GObject      *child,
+						    const gchar  *type);
+static gboolean gtk_container_buildable_custom_tag_start (GtkBuildable  *buildable,
+							  GtkBuilder    *builder,
+							  GObject       *child,
+							  const gchar   *tagname,
+							  GMarkupParser *parser,
+							  gpointer      *data);
+static void    gtk_container_buildable_custom_tag_end (GtkBuildable *buildable,
+						       GtkBuilder   *builder,
+						       GObject      *child,
+						       const gchar  *tagname,
+						       gpointer     *data);
+
 
 /* --- variables --- */
 static const gchar           vadjustment_key[] = "gtk-vadjustment";
@@ -110,6 +129,7 @@ static guint                 container_signals[LAST_SIGNAL] = { 0 };
 static GtkWidgetClass       *parent_class = NULL;
 extern GParamSpecPool       *_gtk_widget_child_property_pool;
 extern GObjectNotifyContext *_gtk_widget_child_property_notify_context;
+static GtkBuildableIface    *parent_buildable_iface;
 
 
 /* --- functions --- */
@@ -120,7 +140,7 @@ gtk_container_get_type (void)
 
   if (!container_type)
     {
-      static const GTypeInfo container_info =
+      const GTypeInfo container_info =
       {
 	sizeof (GtkContainerClass),
 	(GBaseInitFunc) gtk_container_base_class_init,
@@ -134,9 +154,21 @@ gtk_container_get_type (void)
 	NULL,       /* value_table */
       };
 
+      static const GInterfaceInfo buildable_info =
+      {
+	(GInterfaceInitFunc) gtk_container_buildable_init,
+	NULL,
+	NULL
+      };
+
       container_type =
 	g_type_register_static (GTK_TYPE_WIDGET, I_("GtkContainer"), 
 				&container_info, G_TYPE_FLAG_ABSTRACT);
+
+      g_type_add_interface_static (container_type,
+				   GTK_TYPE_BUILDABLE,
+				   &buildable_info);
+
     }
 
   return container_type;
@@ -260,9 +292,172 @@ gtk_container_class_init (GtkContainerClass *class)
 		  GTK_TYPE_WIDGET);
 }
 
+static void
+gtk_container_buildable_init (GtkBuildableIface *iface)
+{
+  parent_buildable_iface = g_type_interface_peek_parent (iface);
+  iface->add_child = gtk_container_buildable_add_child;
+  iface->custom_tag_start = gtk_container_buildable_custom_tag_start;
+  iface->custom_tag_end = gtk_container_buildable_custom_tag_end;
+}
+
+static void
+gtk_container_buildable_add_child (GtkBuildable  *buildable,
+				   GtkBuilder    *builder,
+				   GObject       *child,
+				   const gchar   *type)
+{
+  g_return_if_fail (GTK_IS_WIDGET (child));
+
+  gtk_container_add (GTK_CONTAINER (buildable), GTK_WIDGET (child));
+}
+
+static void
+gtk_container_buildable_set_child_property (GtkContainer *container,
+					    GtkBuilder   *builder,
+					    GtkWidget    *child,
+					    gchar        *name,
+					    const gchar  *value)
+{
+  GParamSpec *pspec;
+  GValue gvalue = { 0, };
+  GError *error = NULL;
+  
+  pspec = gtk_container_class_find_child_property
+    (G_OBJECT_GET_CLASS (container), name);
+  if (!pspec)
+    {
+      g_warning ("%s does not have a property called %s",
+		 g_type_name (G_OBJECT_TYPE (container)), name);
+      return;
+    }
+
+  if (!gtk_builder_value_from_string (builder, pspec, value, &gvalue, &error))
+    {
+      g_warning ("Could not read property %s:%s with value %s of type %s: %s",
+		 g_type_name (G_OBJECT_TYPE (container)),
+		 name,
+		 value,
+		 g_type_name (G_PARAM_SPEC_VALUE_TYPE (pspec)),
+		 error->message);
+      g_error_free (error);
+      return;
+    }
+
+  gtk_container_child_set_property (container, child, name, &gvalue);
+  g_value_unset (&gvalue);
+}
+
+typedef struct {
+  GtkBuilder   *builder;
+  GtkContainer *container;
+  GtkWidget    *child;
+  gchar        *child_prop_name;
+} PackingPropertiesData;
+
+static void
+attributes_start_element (GMarkupParseContext *context,
+			  const gchar         *element_name,
+			  const gchar        **names,
+			  const gchar        **values,
+			  gpointer             user_data,
+			  GError             **error)
+{
+  PackingPropertiesData *parser_data = (PackingPropertiesData*)user_data;
+  guint i;
+
+  if (strcmp (element_name, "property") == 0)
+    for (i = 0; names[i]; i++)
+      if (strcmp (names[i], "name") == 0)
+	parser_data->child_prop_name = g_strdup (values[i]);
+  else if (strcmp (element_name, "packing") == 0)
+    return;
+  else
+    g_warning ("Unsupported tag for GtkContainer: %s\n", element_name);
+}
+
+static void
+attributes_text_element (GMarkupParseContext *context,
+			 const gchar         *text,
+			 gsize                text_len,
+			 gpointer             user_data,
+			 GError             **error)
+{
+  PackingPropertiesData *parser_data = (PackingPropertiesData*)user_data;
+
+  if (!parser_data->child_prop_name)
+    return;
+
+  gtk_container_buildable_set_child_property (parser_data->container,
+					      parser_data->builder,
+					      parser_data->child,
+					      parser_data->child_prop_name,
+					      text);
+
+  g_free (parser_data->child_prop_name);
+  parser_data->child_prop_name = NULL;
+}
+
+static const GMarkupParser attributes_parser =
+  {
+    attributes_start_element,
+    NULL,
+    attributes_text_element,
+  };
+
+static gboolean
+gtk_container_buildable_custom_tag_start (GtkBuildable  *buildable,
+					  GtkBuilder    *builder,
+					  GObject       *child,
+					  const gchar   *tagname,
+					  GMarkupParser *parser,
+					  gpointer      *data)
+{
+  PackingPropertiesData *parser_data;
+
+  if (parent_buildable_iface->custom_tag_start (buildable, builder, child,
+						tagname, parser, data))
+    return TRUE;
+
+  if (child && strcmp (tagname, "packing") == 0)
+    {
+      parser_data = g_slice_new0 (PackingPropertiesData);
+      parser_data->builder = builder;
+      parser_data->container = GTK_CONTAINER (buildable);
+      parser_data->child = GTK_WIDGET (child);
+      parser_data->child_prop_name = NULL;
+
+      *parser = attributes_parser;
+      *data = parser_data;
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static void
+gtk_container_buildable_custom_tag_end (GtkBuildable *buildable,
+					GtkBuilder   *builder,
+					GObject      *child,
+					const gchar  *tagname,
+					gpointer     *data)
+{
+  if (strcmp (tagname, "packing") == 0)
+    {
+      g_slice_free (PackingPropertiesData, (gpointer)data);
+      return;
+
+    }
+
+  if (parent_buildable_iface->custom_tag_end)
+    parent_buildable_iface->custom_tag_end (buildable, builder,
+					    child, tagname, data);
+
+}
+
 /**
  * gtk_container_child_type: 
- * @container: a #GtkContainer.
+ * @container: a #GtkContainer
  *
  * Returns the type of the children supported by the container.
  *
@@ -487,7 +682,7 @@ gtk_container_child_get_property (GtkContainer *container,
  * @child: a widget which is a child of @container
  * @first_property_name: the name of the first property to set
  * @var_args: a %NULL-terminated list of property names and values, starting
- *           with @first_prop_name.
+ *           with @first_prop_name
  * 
  * Sets one or more child properties for @child and @container.
  **/
@@ -610,7 +805,7 @@ gtk_container_child_set_property (GtkContainer *container,
  * @widget: a widget to be placed inside @container 
  * @first_prop_name: the name of the first child property to set 
  * @Varargs: a %NULL-terminated list of property names and values, starting
- *           with @first_prop_name.
+ *           with @first_prop_name
  * 
  * Adds @widget to @container, setting child properties at the same time.
  * See gtk_container_add() and gtk_container_child_set() for more details.
@@ -650,7 +845,7 @@ gtk_container_add_with_properties (GtkContainer *container,
  * @child: a widget which is a child of @container
  * @first_prop_name: the name of the first property to set
  * @Varargs: a %NULL-terminated list of property names and values, starting
- *           with @first_prop_name.
+ *           with @first_prop_name
  * 
  * Sets one or more child properties for @child and @container.
  **/
@@ -677,7 +872,7 @@ gtk_container_child_set (GtkContainer      *container,
  * @child: a widget which is a child of @container
  * @first_prop_name: the name of the first property to get
  * @Varargs: a %NULL-terminated list of property names and #GValue*, 
- *           starting with @first_prop_name.
+ *           starting with @first_prop_name
  * 
  * Gets the values of one or more child properties for @child and @container.
  **/
@@ -761,8 +956,8 @@ gtk_container_class_find_child_property (GObjectClass *cclass,
  * gtk_container_class_list_child_properties:
  * @cclass: a #GtkContainerClass
  * @n_properties: location to return the number of child properties found
- * @returns: a newly allocated array of #GParamSpec*. The array must be 
- *           freed with g_free().
+ * @returns: a newly allocated %NULL-terminated array of #GParamSpec*. 
+ *           The array must be freed with g_free().
  *
  * Returns all child properties of a container class.
  */
@@ -878,8 +1073,8 @@ gtk_container_get_property (GObject         *object,
 /**
  * gtk_container_set_border_width:
  * @container: a #GtkContainer
- * @border_width: amount of blank space to leave <emphasis>outside</emphasis> the container.
- *   Valid values are in the range 0-65535 pixels.
+ * @border_width: amount of blank space to leave <emphasis>outside</emphasis> 
+ *   the container. Valid values are in the range 0-65535 pixels.
  *
  * Sets the border width of the container.
  *
@@ -1003,8 +1198,8 @@ _gtk_container_dequeue_resize_handler (GtkContainer *container)
 
 /**
  * gtk_container_set_resize_mode:
- * @container: a #GtkContainer.
- * @resize_mode: the new resize mode.
+ * @container: a #GtkContainer
+ * @resize_mode: the new resize mode
  * 
  * Sets the resize mode for the container.
  *
@@ -1053,8 +1248,8 @@ gtk_container_get_resize_mode (GtkContainer *container)
 
 /**
  * gtk_container_set_reallocate_redraws:
- * @container: a #GtkContainer.
- * @needs_redraws: the new value for the container's @reallocate_redraws flag.
+ * @container: a #GtkContainer
+ * @needs_redraws: the new value for the container's @reallocate_redraws flag
  *
  * Sets the @reallocate_redraws flag of the container to the given value.
  * 
@@ -1088,8 +1283,6 @@ gtk_container_get_resize_container (GtkContainer *container)
 static gboolean
 gtk_container_idle_sizer (gpointer data)
 {
-  GDK_THREADS_ENTER ();
-
   /* we may be invoked with a container_resize_queue of NULL, because
    * queue_resize could have been adding an extra idle function while
    * the queue still got processed. we better just ignore such case
@@ -1112,8 +1305,6 @@ gtk_container_idle_sizer (gpointer data)
 
   gdk_window_process_all_updates ();
 
-  GDK_THREADS_LEAVE ();
-
   return FALSE;
 }
 
@@ -1151,7 +1342,7 @@ _gtk_container_queue_resize (GtkContainer *container)
 		{
 		  GTK_PRIVATE_SET_FLAG (resize_container, GTK_RESIZE_PENDING);
 		  if (container_resize_queue == NULL)
-		    g_idle_add_full (GTK_PRIORITY_RESIZE,
+		    gdk_threads_add_idle_full (GTK_PRIORITY_RESIZE,
 				     gtk_container_idle_sizer,
 				     NULL, NULL);
 		  container_resize_queue = g_slist_prepend (container_resize_queue, resize_container);
@@ -1267,7 +1458,7 @@ gtk_container_forall (GtkContainer *container,
  * @callback: a callback
  * @callback_data: callback user data
  * 
- * Invokes @callback on each non-internal child of @container.  See
+ * Invokes @callback on each non-internal child of @container. See
  * gtk_container_forall() for details on what constitutes an
  * "internal" child.  Most applications should use
  * gtk_container_foreach(), rather than gtk_container_forall().
@@ -1358,7 +1549,7 @@ gtk_container_set_focus_child (GtkContainer *container,
 
 /**
  * gtk_container_get_children:
- * @container: a #GtkContainer.
+ * @container: a #GtkContainer
  * 
  * Returns the container's non-internal children. See
  * gtk_container_forall() for details on what constitutes an "internal" child. 
@@ -1954,13 +2145,13 @@ gtk_container_focus_sort_left_right (GtkContainer     *container,
  * gtk_container_focus_sort:
  * @container: a #GtkContainer
  * @children:  a list of descendents of @container (they don't
- *             have to be direct children.
+ *             have to be direct children)
  * @direction: focus direction
  * @old_focus: widget to use for the starting position, or %NULL
  *             to determine this automatically.
- *             [ Note, this argument isn't used for GTK_DIR_TAB_*,
- *               which is the only @direction we use currently,
- *               so perhaps this argument should be removed ]
+ *             (Note, this argument isn't used for GTK_DIR_TAB_*,
+ *              which is the only @direction we use currently,
+ *              so perhaps this argument should be removed)
  * 
  * Sorts @children in the correct order for focusing with
  * direction type @direction.
@@ -2077,8 +2268,8 @@ chain_widget_destroyed (GtkWidget *widget,
 
 /**
  * gtk_container_set_focus_chain: 
- * @container: a #GtkContainer.
- * @focusable_widgets: the new focus chain.
+ * @container: a #GtkContainer
+ * @focusable_widgets: the new focus chain
  *
  * Sets a focus chain, overriding the one computed automatically by GTK+.
  * 
@@ -2169,7 +2360,7 @@ gtk_container_get_focus_chain (GtkContainer *container,
 
 /**
  * gtk_container_unset_focus_chain:
- * @container: a #GtkContainer.
+ * @container: a #GtkContainer
  * 
  * Removes a focus chain explicitly set with gtk_container_set_focus_chain().
  **/
@@ -2208,17 +2399,18 @@ gtk_container_unset_focus_chain (GtkContainer  *container)
 /**
  * gtk_container_set_focus_vadjustment:
  * @container: a #GtkContainer
- * @adjustment: an adjustment which should be adjusted when the focus is moved among the
- *   descendents of @container
+ * @adjustment: an adjustment which should be adjusted when the focus 
+ *   is moved among the descendents of @container
  * 
- * Hooks up an adjustment to focus handling in a container, so when a child of the 
- * container is focused, the adjustment is scrolled to show that widget. This function
- * sets the vertical alignment. See gtk_scrolled_window_get_vadjustment() for a typical
- * way of obtaining the adjustment and gtk_container_set_focus_hadjustment() for setting
+ * Hooks up an adjustment to focus handling in a container, so when a 
+ * child of the container is focused, the adjustment is scrolled to 
+ * show that widget. This function sets the vertical alignment. See 
+ * gtk_scrolled_window_get_vadjustment() for a typical way of obtaining 
+ * the adjustment and gtk_container_set_focus_hadjustment() for setting
  * the horizontal adjustment.
  *
- * The adjustments have to be in pixel units and in the same coordinate system as the 
- * allocation for immediate children of the container. 
+ * The adjustments have to be in pixel units and in the same coordinate 
+ * system as the allocation for immediate children of the container. 
  */
 void
 gtk_container_set_focus_vadjustment (GtkContainer  *container,
@@ -2242,7 +2434,7 @@ gtk_container_set_focus_vadjustment (GtkContainer  *container,
  * @container: a #GtkContainer
  *
  * Retrieves the vertical focus adjustment for the container. See
- * gtk_container_set_focus_vadjustment ().
+ * gtk_container_set_focus_vadjustment().
  *
  * Return value: the vertical focus adjustment, or %NULL if
  *   none has been set.
@@ -2262,17 +2454,18 @@ gtk_container_get_focus_vadjustment (GtkContainer *container)
 /**
  * gtk_container_set_focus_hadjustment:
  * @container: a #GtkContainer
- * @adjustment: an adjustment which should be adjusted when the focus is moved among the
- *   descendents of @container
+ * @adjustment: an adjustment which should be adjusted when the focus is 
+ *   moved among the descendents of @container
  * 
- * Hooks up an adjustment to focus handling in a container, so when a child of the 
- * container is focused, the adjustment is scrolled to show that widget. This function
- * sets the horizontal alignment. See gtk_scrolled_window_get_hadjustment() for a typical
- * way of obtaining the adjustment and gtk_container_set_focus_vadjustment() for setting
+ * Hooks up an adjustment to focus handling in a container, so when a child 
+ * of the container is focused, the adjustment is scrolled to show that 
+ * widget. This function sets the horizontal alignment. 
+ * See gtk_scrolled_window_get_hadjustment() for a typical way of obtaining 
+ * the adjustment and gtk_container_set_focus_vadjustment() for setting
  * the vertical adjustment.
  *
- * The adjustments have to be in pixel units and in the same coordinate system as the 
- * allocation for immediate children of the container. 
+ * The adjustments have to be in pixel units and in the same coordinate 
+ * system as the allocation for immediate children of the container. 
  */
 void
 gtk_container_set_focus_hadjustment (GtkContainer  *container,
@@ -2430,7 +2623,7 @@ gtk_container_unmap (GtkWidget *widget)
  * the event's area with the child area, and sending the event.
  * 
  * In most cases, a container can simply either simply inherit the
- * ::expose implementation from #GtkContainer, or, do some drawing 
+ * #GtkWidget::expose implementation from #GtkContainer, or, do some drawing 
  * and then chain to the ::expose implementation from #GtkContainer.
  **/
 void
