@@ -26,6 +26,7 @@
 
 #undef GTK_DISABLE_DEPRECATED
 
+#include <config.h>
 #include <string.h> /* memset */
 
 #include "gtklist.h"
@@ -34,6 +35,8 @@
 #include "gtksignal.h"
 #include "gtklabel.h"
 #include "gtkmarshalers.h"
+#include "gtkintl.h"
+#include "gtkalias.h"
 
 enum {
   SELECTION_CHANGED,
@@ -192,6 +195,7 @@ gtk_list_get_type (void)
         (GtkClassInitFunc) NULL,
       };
 
+      I_("GtkList");
       list_type = gtk_type_unique (GTK_TYPE_CONTAINER, &list_info);
     }
 
@@ -243,14 +247,14 @@ gtk_list_class_init (GtkListClass *class)
   class->unselect_child = gtk_real_list_unselect_child;
 
   list_signals[SELECTION_CHANGED] =
-    gtk_signal_new ("selection_changed",
+    gtk_signal_new (I_("selection-changed"),
 		    GTK_RUN_FIRST,
 		    GTK_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GtkListClass, selection_changed),
 		    _gtk_marshal_VOID__VOID,
 		    GTK_TYPE_NONE, 0);
   list_signals[SELECT_CHILD] =
-    gtk_signal_new ("select_child",
+    gtk_signal_new (I_("select_child"),
 		    GTK_RUN_FIRST,
 		    GTK_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GtkListClass, select_child),
@@ -258,7 +262,7 @@ gtk_list_class_init (GtkListClass *class)
 		    GTK_TYPE_NONE, 1,
 		    GTK_TYPE_WIDGET);
   list_signals[UNSELECT_CHILD] =
-    gtk_signal_new ("unselect_child",
+    gtk_signal_new (I_("unselect_child"),
 		    GTK_RUN_FIRST,
 		    GTK_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GtkListClass, unselect_child),
@@ -266,8 +270,9 @@ gtk_list_class_init (GtkListClass *class)
 		    GTK_TYPE_NONE, 1,
 		    GTK_TYPE_WIDGET);
   
-  gtk_object_add_arg_type ("GtkList::selection_mode",
-			   GTK_TYPE_SELECTION_MODE, GTK_ARG_READWRITE,
+  gtk_object_add_arg_type ("GtkList::selection-mode",
+			   GTK_TYPE_SELECTION_MODE, 
+			   GTK_ARG_READWRITE | G_PARAM_STATIC_NAME,
 			   ARG_SELECTION_MODE);
 }
 
@@ -552,8 +557,8 @@ gtk_list_motion_notify (GtkWidget      *widget,
     {
       if (list->htimer == 0)
 	{
-	  list->htimer = gtk_timeout_add
-	    (SCROLL_TIME, (GtkFunction) gtk_list_horizontal_timeout, widget);
+	  list->htimer = gdk_threads_add_timeout
+	    (SCROLL_TIME, (GSourceFunc) gtk_list_horizontal_timeout, widget);
 	  
 	  if (!((x < adj->value && adj->value <= 0) ||
 		(x > adj->value + adj->page_size &&
@@ -600,9 +605,9 @@ gtk_list_motion_notify (GtkWidget      *widget,
 
   if (!((y < 0 && focus_row == 0) ||
 	(y > widget->allocation.height && focus_row >= length - 1)))
-    list->vtimer = gtk_timeout_add (SCROLL_TIME,
-				    (GtkFunction) gtk_list_vertical_timeout,
-				    list);
+    list->vtimer = gdk_threads_add_timeout (SCROLL_TIME,
+				  (GSourceFunc) gtk_list_vertical_timeout,
+				  list);
 
   if (row != focus_row)
     gtk_widget_grab_focus (item);
@@ -1591,12 +1596,12 @@ gtk_list_end_drag_selection (GtkList *list)
 
   if (list->htimer)
     {
-      gtk_timeout_remove (list->htimer);
+      g_source_remove (list->htimer);
       list->htimer = 0;
     }
   if (list->vtimer)
     {
-      gtk_timeout_remove (list->vtimer);
+      g_source_remove (list->vtimer);
       list->vtimer = 0;
     }
 }
@@ -2333,26 +2338,18 @@ do_fake_motion (GtkWidget *list)
 static gint
 gtk_list_horizontal_timeout (GtkWidget *list)
 {
-  GDK_THREADS_ENTER ();
-
   GTK_LIST (list)->htimer = 0;
   do_fake_motion (list);
   
-  GDK_THREADS_LEAVE ();
-
   return FALSE;
 }
 
 static gint
 gtk_list_vertical_timeout (GtkWidget *list)
 {
-  GDK_THREADS_ENTER ();
-
   GTK_LIST (list)->vtimer = 0;
   do_fake_motion (list);
 
-  GDK_THREADS_LEAVE ();
-
   return FALSE;
 }
 
@@ -2616,3 +2613,6 @@ gtk_list_drag_begin (GtkWidget      *widget,
 	}
     }
 }
+
+#define __GTK_LIST_C__
+#include "gtkaliasdef.c"
