@@ -18,23 +18,26 @@
  */
 
 #include <config.h>
+#include <errno.h>
+#include <stdlib.h>
 #include <string.h>
 #include <gobject/gvaluecollector.h>
 #include "gtktreemodel.h"
 #include "gtkliststore.h"
 #include "gtktreedatalist.h"
 #include "gtktreednd.h"
-#include "gtksequence.h"
 #include "gtkintl.h"
+#include "gtkbuildable.h"
 #include "gtkalias.h"
 
 #define GTK_LIST_STORE_IS_SORTED(list) (((GtkListStore*)(list))->sort_column_id != GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID)
-#define VALID_ITER(iter, list_store) ((iter)!= NULL && (iter)->user_data != NULL && list_store->stamp == (iter)->stamp && !_gtk_sequence_ptr_is_end ((iter)->user_data) && _gtk_sequence_ptr_get_sequence ((iter)->user_data) == list_store->seq)
+#define VALID_ITER(iter, list_store) ((iter)!= NULL && (iter)->user_data != NULL && list_store->stamp == (iter)->stamp && !g_sequence_iter_is_end ((iter)->user_data) && g_sequence_iter_get_sequence ((iter)->user_data) == list_store->seq)
 
 static void         gtk_list_store_tree_model_init (GtkTreeModelIface *iface);
 static void         gtk_list_store_drag_source_init(GtkTreeDragSourceIface *iface);
 static void         gtk_list_store_drag_dest_init  (GtkTreeDragDestIface   *iface);
 static void         gtk_list_store_sortable_init   (GtkTreeSortableIface   *iface);
+static void         gtk_list_store_buildable_init  (GtkBuildableIface      *iface);
 static void         gtk_list_store_finalize        (GObject           *object);
 static GtkTreeModelFlags gtk_list_store_get_flags  (GtkTreeModel      *tree_model);
 static gint         gtk_list_store_get_n_columns   (GtkTreeModel      *tree_model);
@@ -115,6 +118,19 @@ static void     gtk_list_store_set_default_sort_func (GtkTreeSortable        *so
 static gboolean gtk_list_store_has_default_sort_func (GtkTreeSortable        *sortable);
 
 
+/* buildable */
+static gboolean gtk_list_store_buildable_custom_tag_start (GtkBuildable  *buildable,
+							   GtkBuilder    *builder,
+							   GObject       *child,
+							   const gchar   *tagname,
+							   GMarkupParser *parser,
+							   gpointer      *data);
+static void     gtk_list_store_buildable_custom_tag_end (GtkBuildable *buildable,
+							 GtkBuilder   *builder,
+							 GObject      *child,
+							 const gchar  *tagname,
+							 gpointer     *data);
+
 G_DEFINE_TYPE_WITH_CODE (GtkListStore, gtk_list_store, G_TYPE_OBJECT,
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_MODEL,
 						gtk_list_store_tree_model_init)
@@ -123,7 +139,10 @@ G_DEFINE_TYPE_WITH_CODE (GtkListStore, gtk_list_store, G_TYPE_OBJECT,
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_DRAG_DEST,
 						gtk_list_store_drag_dest_init)
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_TREE_SORTABLE,
-						gtk_list_store_sortable_init))
+						gtk_list_store_sortable_init)
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+						gtk_list_store_buildable_init))
+
 
 static void
 gtk_list_store_class_init (GtkListStoreClass *class)
@@ -177,10 +196,17 @@ gtk_list_store_sortable_init (GtkTreeSortableIface *iface)
   iface->has_default_sort_func = gtk_list_store_has_default_sort_func;
 }
 
+void
+gtk_list_store_buildable_init (GtkBuildableIface *iface)
+{
+  iface->custom_tag_start = gtk_list_store_buildable_custom_tag_start;
+  iface->custom_tag_end = gtk_list_store_buildable_custom_tag_end;
+}
+
 static void
 gtk_list_store_init (GtkListStore *list_store)
 {
-  list_store->seq = _gtk_sequence_new (NULL);
+  list_store->seq = g_sequence_new (NULL);
   list_store->sort_list = NULL;
   list_store->stamp = g_random_int ();
   list_store->sort_column_id = -2;
@@ -295,7 +321,7 @@ gtk_list_store_set_column_types (GtkListStore *list_store,
   g_return_if_fail (list_store->columns_dirty == 0);
 
   gtk_list_store_set_n_columns (list_store, n_columns);
-   for (i = 0; i < n_columns; i++)
+  for (i = 0; i < n_columns; i++)
     {
       if (! _gtk_tree_data_list_check_type (types[i]))
 	{
@@ -355,10 +381,10 @@ gtk_list_store_finalize (GObject *object)
 {
   GtkListStore *list_store = GTK_LIST_STORE (object);
 
-  _gtk_sequence_foreach (list_store->seq,
-			(GFunc) _gtk_tree_data_list_free, list_store->column_headers);
+  g_sequence_foreach (list_store->seq,
+		      (GFunc) _gtk_tree_data_list_free, list_store->column_headers);
 
-  _gtk_sequence_free (list_store->seq);
+  g_sequence_free (list_store->seq);
 
   _gtk_tree_data_list_header_free (list_store->sort_list);
   g_free (list_store->column_headers);
@@ -413,7 +439,7 @@ gtk_list_store_get_iter (GtkTreeModel *tree_model,
 			 GtkTreePath  *path)
 {
   GtkListStore *list_store = (GtkListStore *) tree_model;
-  GtkSequence *seq;
+  GSequence *seq;
   gint i;
 
   list_store->columns_dirty = TRUE;
@@ -422,11 +448,11 @@ gtk_list_store_get_iter (GtkTreeModel *tree_model,
   
   i = gtk_tree_path_get_indices (path)[0];
 
-  if (i >= _gtk_sequence_get_length (seq))
+  if (i >= g_sequence_get_length (seq))
     return FALSE;
 
   iter->stamp = list_store->stamp;
-  iter->user_data = _gtk_sequence_get_ptr_at_pos (seq, i);
+  iter->user_data = g_sequence_get_iter_at_pos (seq, i);
 
   return TRUE;
 }
@@ -439,11 +465,11 @@ gtk_list_store_get_path (GtkTreeModel *tree_model,
 
   g_return_val_if_fail (iter->stamp == GTK_LIST_STORE (tree_model)->stamp, NULL);
 
-  if (_gtk_sequence_ptr_is_end (iter->user_data))
+  if (g_sequence_iter_is_end (iter->user_data))
     return NULL;
 	
   path = gtk_tree_path_new ();
-  gtk_tree_path_append_index (path, _gtk_sequence_ptr_get_position (iter->user_data));
+  gtk_tree_path_append_index (path, g_sequence_iter_get_position (iter->user_data));
   
   return path;
 }
@@ -461,7 +487,7 @@ gtk_list_store_get_value (GtkTreeModel *tree_model,
   g_return_if_fail (column < list_store->n_columns);
   g_return_if_fail (VALID_ITER (iter, list_store));
 		    
-  list = _gtk_sequence_ptr_get_data (iter->user_data);
+  list = g_sequence_get (iter->user_data);
 
   while (tmp_column-- > 0 && list)
     list = list->next;
@@ -479,9 +505,9 @@ gtk_list_store_iter_next (GtkTreeModel  *tree_model,
 			  GtkTreeIter   *iter)
 {
   g_return_val_if_fail (GTK_LIST_STORE (tree_model)->stamp == iter->stamp, FALSE);
-  iter->user_data = _gtk_sequence_ptr_next (iter->user_data);
+  iter->user_data = g_sequence_iter_next (iter->user_data);
 
-  return !_gtk_sequence_ptr_is_end (iter->user_data);
+  return !g_sequence_iter_is_end (iter->user_data);
 }
 
 static gboolean
@@ -495,10 +521,10 @@ gtk_list_store_iter_children (GtkTreeModel *tree_model,
   if (parent)
     return FALSE;
 
-  if (_gtk_sequence_get_length (list_store->seq) > 0)
+  if (g_sequence_get_length (list_store->seq) > 0)
     {
       iter->stamp = list_store->stamp;
-      iter->user_data = _gtk_sequence_get_begin_ptr (list_store->seq);
+      iter->user_data = g_sequence_get_begin_iter (list_store->seq);
       return TRUE;
     }
   else
@@ -519,7 +545,7 @@ gtk_list_store_iter_n_children (GtkTreeModel *tree_model,
   GtkListStore *list_store = (GtkListStore *) tree_model;
 
   if (iter == NULL)
-    return _gtk_sequence_get_length (list_store->seq);
+    return g_sequence_get_length (list_store->seq);
 
   g_return_val_if_fail (list_store->stamp == iter->stamp, -1);
 
@@ -533,14 +559,14 @@ gtk_list_store_iter_nth_child (GtkTreeModel *tree_model,
 			       gint          n)
 {
   GtkListStore *list_store = (GtkListStore *) tree_model;
-  GtkSequencePtr child;
+  GSequenceIter *child;
 
   if (parent)
     return FALSE;
 
-  child = _gtk_sequence_get_ptr_at_pos (list_store->seq, n);
+  child = g_sequence_get_iter_at_pos (list_store->seq, n);
 
-  if (_gtk_sequence_ptr_is_end (child))
+  if (g_sequence_iter_is_end (child))
     return FALSE;
 
   iter->stamp = list_store->stamp;
@@ -594,7 +620,7 @@ gtk_list_store_real_set_value (GtkListStore *list_store,
       converted = TRUE;
     }
 
-  prev = list = _gtk_sequence_ptr_get_data (iter->user_data);
+  prev = list = g_sequence_get (iter->user_data);
 
   while (list != NULL)
     {
@@ -617,10 +643,10 @@ gtk_list_store_real_set_value (GtkListStore *list_store,
       list = list->next;
     }
 
-  if (_gtk_sequence_ptr_get_data (iter->user_data) == NULL)
+  if (g_sequence_get (iter->user_data) == NULL)
     {
       list = _gtk_tree_data_list_alloc();
-      _gtk_sequence_set (iter->user_data, list);
+      g_sequence_set (iter->user_data, list);
       list->next = NULL;
     }
   else
@@ -711,6 +737,36 @@ gtk_list_store_get_compare_func (GtkListStore *list_store)
   return func;
 }
 
+static void
+gtk_list_store_set_vector_internal (GtkListStore *list_store,
+				    GtkTreeIter  *iter,
+				    gboolean     *emit_signal,
+				    gboolean     *maybe_need_sort,
+				    gint         *columns,
+				    GValue       *values,
+				    gint          n_values)
+{
+  gint i;
+  GtkTreeIterCompareFunc func = NULL;
+
+  func = gtk_list_store_get_compare_func (list_store);
+  if (func != _gtk_tree_data_list_compare_func)
+    *maybe_need_sort = TRUE;
+
+  for (i = 0; i < n_values; i++)
+    {
+      *emit_signal = gtk_list_store_real_set_value (list_store, 
+					       iter, 
+					       columns[i],
+					       &values[i],
+					       FALSE) || *emit_signal;
+
+      if (func == _gtk_tree_data_list_compare_func &&
+	  columns[i] == list_store->sort_column_id)
+	*maybe_need_sort = TRUE;
+    }
+}
+
 static void
 gtk_list_store_set_valist_internal (GtkListStore *list_store,
 				    GtkTreeIter  *iter,
@@ -768,6 +824,53 @@ gtk_list_store_set_valist_internal (GtkListStore *list_store,
     }
 }
 
+/**
+ * gtk_list_store_set_valuesv:
+ * @list_store: A #GtkListStore
+ * @iter: A valid #GtkTreeIter for the row being modified
+ * @columns: an array of column numbers
+ * @values: an array of GValues 
+ * @n_values: the length of the @columns and @values arrays
+ * 
+ * A variant of gtk_list_store_set_valist() which
+ * takes the columns and values as two arrays, instead of
+ * varargs. This function is mainly intended for 
+ * language-bindings and in case the number of columns to
+ * change is not known until run-time.
+ *
+ * Since: 2.12
+ */
+void
+gtk_list_store_set_valuesv (GtkListStore *list_store,
+			    GtkTreeIter  *iter,
+			    gint         *columns,
+			    GValue       *values,
+			    gint          n_values)
+{
+  gboolean emit_signal = FALSE;
+  gboolean maybe_need_sort = FALSE;
+
+  g_return_if_fail (GTK_IS_LIST_STORE (list_store));
+  g_return_if_fail (VALID_ITER (iter, list_store));
+
+  gtk_list_store_set_vector_internal (list_store, iter,
+				      &emit_signal,
+				      &maybe_need_sort,
+				      columns, values, n_values);
+
+  if (maybe_need_sort && GTK_LIST_STORE_IS_SORTED (list_store))
+    gtk_list_store_sort_iter_changed (list_store, iter, list_store->sort_column_id);
+
+  if (emit_signal)
+    {
+      GtkTreePath *path;
+
+      path = gtk_list_store_get_path (GTK_TREE_MODEL (list_store), iter);
+      gtk_tree_model_row_changed (GTK_TREE_MODEL (list_store), path, iter);
+      gtk_tree_path_free (path);
+    }
+}
+
 /**
  * gtk_list_store_set_valist:
  * @list_store: A #GtkListStore
@@ -848,7 +951,7 @@ gtk_list_store_remove (GtkListStore *list_store,
 		       GtkTreeIter  *iter)
 {
   GtkTreePath *path;
-  GtkSequencePtr ptr, next;
+  GSequenceIter *ptr, *next;
 
   g_return_val_if_fail (GTK_IS_LIST_STORE (list_store), FALSE);
   g_return_val_if_fail (VALID_ITER (iter, list_store), FALSE);
@@ -856,17 +959,17 @@ gtk_list_store_remove (GtkListStore *list_store,
   path = gtk_list_store_get_path (GTK_TREE_MODEL (list_store), iter);
 
   ptr = iter->user_data;
-  next = _gtk_sequence_ptr_next (ptr);
+  next = g_sequence_iter_next (ptr);
   
-  _gtk_tree_data_list_free (_gtk_sequence_ptr_get_data (ptr), list_store->column_headers);
-  _gtk_sequence_remove (iter->user_data);
+  _gtk_tree_data_list_free (g_sequence_get (ptr), list_store->column_headers);
+  g_sequence_remove (iter->user_data);
 
   list_store->length--;
   
   gtk_tree_model_row_deleted (GTK_TREE_MODEL (list_store), path);
   gtk_tree_path_free (path);
 
-  if (_gtk_sequence_ptr_is_end (next))
+  if (g_sequence_iter_is_end (next))
     {
       iter->stamp = 0;
       return FALSE;
@@ -898,8 +1001,8 @@ gtk_list_store_insert (GtkListStore *list_store,
 		       gint          position)
 {
   GtkTreePath *path;
-  GtkSequence *seq;
-  GtkSequencePtr ptr;
+  GSequence *seq;
+  GSequenceIter *ptr;
   gint length;
 
   g_return_if_fail (GTK_IS_LIST_STORE (list_store));
@@ -910,12 +1013,12 @@ gtk_list_store_insert (GtkListStore *list_store,
 
   seq = list_store->seq;
 
-  length = _gtk_sequence_get_length (seq);
+  length = g_sequence_get_length (seq);
   if (position > length)
     position = length;
 
-  ptr = _gtk_sequence_get_ptr_at_pos (seq, position);
-  ptr = _gtk_sequence_insert (ptr, NULL);
+  ptr = g_sequence_get_iter_at_pos (seq, position);
+  ptr = g_sequence_insert_before (ptr, NULL);
 
   iter->stamp = list_store->stamp;
   iter->user_data = ptr;
@@ -947,7 +1050,7 @@ gtk_list_store_insert_before (GtkListStore *list_store,
 			      GtkTreeIter  *iter,
 			      GtkTreeIter  *sibling)
 {
-  GtkSequencePtr after;
+  GSequenceIter *after;
   
   g_return_if_fail (GTK_IS_LIST_STORE (list_store));
   g_return_if_fail (iter != NULL);
@@ -955,11 +1058,11 @@ gtk_list_store_insert_before (GtkListStore *list_store,
     g_return_if_fail (VALID_ITER (sibling, list_store));
 
   if (!sibling)
-    after = _gtk_sequence_get_end_ptr (list_store->seq);
+    after = g_sequence_get_end_iter (list_store->seq);
   else
     after = sibling->user_data;
 
-  gtk_list_store_insert (list_store, iter, _gtk_sequence_ptr_get_position (after));
+  gtk_list_store_insert (list_store, iter, g_sequence_iter_get_position (after));
 }
 
 /**
@@ -979,7 +1082,7 @@ gtk_list_store_insert_after (GtkListStore *list_store,
 			     GtkTreeIter  *iter,
 			     GtkTreeIter  *sibling)
 {
-  GtkSequencePtr after;
+  GSequenceIter *after;
 
   g_return_if_fail (GTK_IS_LIST_STORE (list_store));
   g_return_if_fail (iter != NULL);
@@ -987,11 +1090,11 @@ gtk_list_store_insert_after (GtkListStore *list_store,
     g_return_if_fail (VALID_ITER (sibling, list_store));
 
   if (!sibling)
-    after = _gtk_sequence_get_begin_ptr (list_store->seq);
+    after = g_sequence_get_begin_iter (list_store->seq);
   else
-    after = _gtk_sequence_ptr_next (sibling->user_data);
+    after = g_sequence_iter_next (sibling->user_data);
 
-  gtk_list_store_insert (list_store, iter, _gtk_sequence_ptr_get_position (after));
+  gtk_list_store_insert (list_store, iter, g_sequence_iter_get_position (after));
 }
 
 /**
@@ -1031,7 +1134,7 @@ gtk_list_store_append (GtkListStore *list_store,
   g_return_if_fail (GTK_IS_LIST_STORE (list_store));
   g_return_if_fail (iter != NULL);
 
-  gtk_list_store_insert (list_store, iter, _gtk_sequence_get_length (list_store->seq));
+  gtk_list_store_insert (list_store, iter, g_sequence_get_length (list_store->seq));
 }
 
 static void
@@ -1057,10 +1160,10 @@ gtk_list_store_clear (GtkListStore *list_store)
   GtkTreeIter iter;
   g_return_if_fail (GTK_IS_LIST_STORE (list_store));
 
-  while (_gtk_sequence_get_length (list_store->seq) > 0)
+  while (g_sequence_get_length (list_store->seq) > 0)
     {
       iter.stamp = list_store->stamp;
-      iter.user_data = _gtk_sequence_get_begin_ptr (list_store->seq);
+      iter.user_data = g_sequence_get_begin_iter (list_store->seq);
       gtk_list_store_remove (list_store, &iter);
     }
 
@@ -1091,7 +1194,7 @@ gtk_list_store_iter_is_valid (GtkListStore *list_store,
   if (!VALID_ITER (iter, list_store))
     return FALSE;
 
-  if (_gtk_sequence_ptr_get_sequence (iter->user_data) != list_store->seq)
+  if (g_sequence_iter_get_sequence (iter->user_data) != list_store->seq)
     return FALSE;
 
   return TRUE;
@@ -1205,7 +1308,7 @@ gtk_list_store_drag_data_received (GtkTreeDragDest   *drag_dest,
        */
       if (retval)
         {
-          GtkTreeDataList *dl = _gtk_sequence_ptr_get_data (src_iter.user_data);
+          GtkTreeDataList *dl = g_sequence_get (src_iter.user_data);
           GtkTreeDataList *copy_head = NULL;
           GtkTreeDataList *copy_prev = NULL;
           GtkTreeDataList *copy_iter = NULL;
@@ -1231,7 +1334,7 @@ gtk_list_store_drag_data_received (GtkTreeDragDest   *drag_dest,
             }
 
 	  dest_iter.stamp = list_store->stamp;
-          _gtk_sequence_set (dest_iter.user_data, copy_head);
+          g_sequence_set (dest_iter.user_data, copy_head);
 
 	  path = gtk_list_store_get_path (tree_model, &dest_iter);
 	  gtk_tree_model_row_changed (tree_model, path, &dest_iter);
@@ -1282,7 +1385,7 @@ gtk_list_store_row_drop_possible (GtkTreeDragDest  *drag_dest,
 
   indices = gtk_tree_path_get_indices (dest_path);
 
-  if (indices[0] <= _gtk_sequence_get_length (GTK_LIST_STORE (drag_dest)->seq))
+  if (indices[0] <= g_sequence_get_length (GTK_LIST_STORE (drag_dest)->seq))
     retval = TRUE;
 
  out:
@@ -1296,9 +1399,9 @@ gtk_list_store_row_drop_possible (GtkTreeDragDest  *drag_dest,
 
 /* Reordering */
 static gint
-gtk_list_store_reorder_func (gconstpointer a,
-			     gconstpointer b,
-			     gpointer      user_data)
+gtk_list_store_reorder_func (GSequenceIter *a,
+			     GSequenceIter *b,
+			     gpointer       user_data)
 {
   GHashTable *new_positions = user_data;
   gint apos = GPOINTER_TO_INT (g_hash_table_lookup (new_positions, a));
@@ -1330,30 +1433,30 @@ gtk_list_store_reorder (GtkListStore *store,
   gint i;
   GtkTreePath *path;
   GHashTable *new_positions;
-  GtkSequencePtr ptr;
+  GSequenceIter *ptr;
   gint *order;
   
   g_return_if_fail (GTK_IS_LIST_STORE (store));
   g_return_if_fail (!GTK_LIST_STORE_IS_SORTED (store));
   g_return_if_fail (new_order != NULL);
 
-  order = g_new (gint, _gtk_sequence_get_length (store->seq));
-  for (i = 0; i < _gtk_sequence_get_length (store->seq); i++)
+  order = g_new (gint, g_sequence_get_length (store->seq));
+  for (i = 0; i < g_sequence_get_length (store->seq); i++)
     order[new_order[i]] = i;
   
   new_positions = g_hash_table_new (g_direct_hash, g_direct_equal);
 
-  ptr = _gtk_sequence_get_begin_ptr (store->seq);
+  ptr = g_sequence_get_begin_iter (store->seq);
   i = 0;
-  while (!_gtk_sequence_ptr_is_end (ptr))
+  while (!g_sequence_iter_is_end (ptr))
     {
       g_hash_table_insert (new_positions, ptr, GINT_TO_POINTER (order[i++]));
 
-      ptr = _gtk_sequence_ptr_next (ptr);
+      ptr = g_sequence_iter_next (ptr);
     }
   g_free (order);
   
-  _gtk_sequence_sort (store->seq, gtk_list_store_reorder_func, new_positions);
+  g_sequence_sort_iter (store->seq, gtk_list_store_reorder_func, new_positions);
 
   g_hash_table_destroy (new_positions);
   
@@ -1365,37 +1468,37 @@ gtk_list_store_reorder (GtkListStore *store,
 }
 
 static GHashTable *
-save_positions (GtkSequence *seq)
+save_positions (GSequence *seq)
 {
   GHashTable *positions = g_hash_table_new (g_direct_hash, g_direct_equal);
-  GtkSequencePtr ptr;
+  GSequenceIter *ptr;
 
-  ptr = _gtk_sequence_get_begin_ptr (seq);
-  while (!_gtk_sequence_ptr_is_end (ptr))
+  ptr = g_sequence_get_begin_iter (seq);
+  while (!g_sequence_iter_is_end (ptr))
     {
       g_hash_table_insert (positions, ptr,
-			   GINT_TO_POINTER (_gtk_sequence_ptr_get_position (ptr)));
-      ptr = _gtk_sequence_ptr_next (ptr);
+			   GINT_TO_POINTER (g_sequence_iter_get_position (ptr)));
+      ptr = g_sequence_iter_next (ptr);
     }
 
   return positions;
 }
 
 static int *
-generate_order (GtkSequence *seq,
+generate_order (GSequence *seq,
 		GHashTable *old_positions)
 {
-  GtkSequencePtr ptr;
-  int *order = g_new (int, _gtk_sequence_get_length (seq));
+  GSequenceIter *ptr;
+  int *order = g_new (int, g_sequence_get_length (seq));
   int i;
 
   i = 0;
-  ptr = _gtk_sequence_get_begin_ptr (seq);
-  while (!_gtk_sequence_ptr_is_end (ptr))
+  ptr = g_sequence_get_begin_iter (seq);
+  while (!g_sequence_iter_is_end (ptr))
     {
       int old_pos = GPOINTER_TO_INT (g_hash_table_lookup (old_positions, ptr));
       order[i++] = old_pos;
-      ptr = _gtk_sequence_ptr_next (ptr);
+      ptr = g_sequence_iter_next (ptr);
     }
 
   g_hash_table_destroy (old_positions);
@@ -1433,7 +1536,7 @@ gtk_list_store_swap (GtkListStore *store,
 
   old_positions = save_positions (store->seq);
   
-  _gtk_sequence_swap (a->user_data, b->user_data);
+  g_sequence_swap (a->user_data, b->user_data);
 
   order = generate_order (store->seq, old_positions);
   path = gtk_tree_path_new ();
@@ -1456,7 +1559,7 @@ gtk_list_store_move_to (GtkListStore *store,
   
   old_positions = save_positions (store->seq);
   
-  _gtk_sequence_move (iter->user_data, _gtk_sequence_get_ptr_at_pos (store->seq, new_pos));
+  g_sequence_move (iter->user_data, g_sequence_get_iter_at_pos (store->seq, new_pos));
 
   order = generate_order (store->seq, old_positions);
   
@@ -1493,7 +1596,7 @@ gtk_list_store_move_before (GtkListStore *store,
     g_return_if_fail (VALID_ITER (position, store));
 
   if (position)
-    pos = _gtk_sequence_ptr_get_position (position->user_data);
+    pos = g_sequence_iter_get_position (position->user_data);
   else
     pos = -1;
   
@@ -1526,7 +1629,7 @@ gtk_list_store_move_after (GtkListStore *store,
     g_return_if_fail (VALID_ITER (position, store));
 
   if (position)
-    pos = _gtk_sequence_ptr_get_position (position->user_data) + 1;
+    pos = g_sequence_iter_get_position (position->user_data) + 1;
   else
     pos = 0;
   
@@ -1535,8 +1638,8 @@ gtk_list_store_move_after (GtkListStore *store,
     
 /* Sorting */
 static gint
-gtk_list_store_compare_func (gconstpointer a,
-			     gconstpointer b,
+gtk_list_store_compare_func (GSequenceIter *a,
+			     GSequenceIter *b,
 			     gpointer      user_data)
 {
   GtkListStore *list_store = user_data;
@@ -1594,12 +1697,12 @@ gtk_list_store_sort (GtkListStore *list_store)
   GHashTable *old_positions;
 
   if (!GTK_LIST_STORE_IS_SORTED (list_store) ||
-      _gtk_sequence_get_length (list_store->seq) <= 1)
+      g_sequence_get_length (list_store->seq) <= 1)
     return;
 
   old_positions = save_positions (list_store->seq);
 
-  _gtk_sequence_sort (list_store->seq, gtk_list_store_compare_func, list_store);
+  g_sequence_sort_iter (list_store->seq, gtk_list_store_compare_func, list_store);
 
   /* Let the world know about our new order */
   new_order = generate_order (list_store->seq, old_positions);
@@ -1615,17 +1718,17 @@ static gboolean
 iter_is_sorted (GtkListStore *list_store,
                 GtkTreeIter  *iter)
 {
-  GtkSequencePtr cmp;
+  GSequenceIter *cmp;
 
-  if (!_gtk_sequence_ptr_is_begin (iter->user_data))
+  if (!g_sequence_iter_is_begin (iter->user_data))
     {
-      cmp = _gtk_sequence_ptr_prev (iter->user_data);
+      cmp = g_sequence_iter_prev (iter->user_data);
       if (gtk_list_store_compare_func (cmp, iter->user_data, list_store) > 0)
 	return FALSE;
     }
 
-  cmp = _gtk_sequence_ptr_next (iter->user_data);
-  if (!_gtk_sequence_ptr_is_end (cmp))
+  cmp = g_sequence_iter_next (iter->user_data);
+  if (!g_sequence_iter_is_end (cmp))
     {
       if (gtk_list_store_compare_func (iter->user_data, cmp, list_store) > 0)
 	return FALSE;
@@ -1652,9 +1755,9 @@ gtk_list_store_sort_iter_changed (GtkListStore *list_store,
       gint *order;
 
       old_positions = save_positions (list_store->seq);
-      _gtk_sequence_sort_changed (iter->user_data,
-                                  gtk_list_store_compare_func,
-                                  list_store);
+      g_sequence_sort_changed_iter (iter->user_data,
+				    gtk_list_store_compare_func,
+				    list_store);
       order = generate_order (list_store->seq, old_positions);
       path = gtk_tree_path_new ();
       gtk_tree_model_rows_reordered (GTK_TREE_MODEL (list_store),
@@ -1807,8 +1910,8 @@ gtk_list_store_insert_with_values (GtkListStore *list_store,
 				   ...)
 {
   GtkTreePath *path;
-  GtkSequence *seq;
-  GtkSequencePtr ptr;
+  GSequence *seq;
+  GSequenceIter *ptr;
   GtkTreeIter tmp_iter;
   gint length;
   gboolean changed = FALSE;
@@ -1825,12 +1928,12 @@ gtk_list_store_insert_with_values (GtkListStore *list_store,
 
   seq = list_store->seq;
 
-  length = _gtk_sequence_get_length (seq);
+  length = g_sequence_get_length (seq);
   if (position > length)
     position = length;
 
-  ptr = _gtk_sequence_get_ptr_at_pos (seq, position);
-  ptr = _gtk_sequence_insert (ptr, NULL);
+  ptr = g_sequence_get_iter_at_pos (seq, position);
+  ptr = g_sequence_insert_before (ptr, NULL);
 
   iter->stamp = list_store->stamp;
   iter->user_data = ptr;
@@ -1847,9 +1950,9 @@ gtk_list_store_insert_with_values (GtkListStore *list_store,
 
   /* Don't emit rows_reordered here */
   if (maybe_need_sort && GTK_LIST_STORE_IS_SORTED (list_store))
-    _gtk_sequence_sort_changed (iter->user_data,
-				gtk_list_store_compare_func,
-				list_store);
+    g_sequence_sort_changed_iter (iter->user_data,
+				  gtk_list_store_compare_func,
+				  list_store);
 
   /* Just emit row_inserted */
   path = gtk_list_store_get_path (GTK_TREE_MODEL (list_store), iter);
@@ -1883,14 +1986,12 @@ gtk_list_store_insert_with_valuesv (GtkListStore *list_store,
 				    gint          n_values)
 {
   GtkTreePath *path;
-  GtkSequence *seq;
-  GtkSequencePtr ptr;
+  GSequence *seq;
+  GSequenceIter *ptr;
   GtkTreeIter tmp_iter;
   gint length;
   gboolean changed = FALSE;
   gboolean maybe_need_sort = FALSE;
-  GtkTreeIterCompareFunc func = NULL;
-  gint i;
 
   /* FIXME refactor to reduce overlap with 
    * gtk_list_store_insert_with_values() 
@@ -1904,12 +2005,12 @@ gtk_list_store_insert_with_valuesv (GtkListStore *list_store,
 
   seq = list_store->seq;
 
-  length = _gtk_sequence_get_length (seq);
+  length = g_sequence_get_length (seq);
   if (position > length)
     position = length;
 
-  ptr = _gtk_sequence_get_ptr_at_pos (seq, position);
-  ptr = _gtk_sequence_insert (ptr, NULL);
+  ptr = g_sequence_get_iter_at_pos (seq, position);
+  ptr = g_sequence_insert_before (ptr, NULL);
 
   iter->stamp = list_store->stamp;
   iter->user_data = ptr;
@@ -1918,28 +2019,15 @@ gtk_list_store_insert_with_valuesv (GtkListStore *list_store,
 
   list_store->length++;  
 
-  func = gtk_list_store_get_compare_func (list_store);
-  if (func != _gtk_tree_data_list_compare_func)
-    maybe_need_sort = TRUE;
-
-  for (i = 0; i < n_values; i++)
-    {
-      changed = gtk_list_store_real_set_value (list_store, 
-					       iter, 
-					       columns[i],
-					       &values[i],
-					       FALSE) || changed;
-
-      if (func == _gtk_tree_data_list_compare_func &&
-	  columns[i] == list_store->sort_column_id)
-	maybe_need_sort = TRUE;
-    }
+  gtk_list_store_set_vector_internal (list_store, iter,
+				      &changed, &maybe_need_sort,
+				      columns, values, n_values);
 
   /* Don't emit rows_reordered here */
   if (maybe_need_sort && GTK_LIST_STORE_IS_SORTED (list_store))
-    _gtk_sequence_sort_changed (iter->user_data,
-				gtk_list_store_compare_func,
-				list_store);
+    g_sequence_sort_changed_iter (iter->user_data,
+				  gtk_list_store_compare_func,
+				  list_store);
 
   /* Just emit row_inserted */
   path = gtk_list_store_get_path (GTK_TREE_MODEL (list_store), iter);
@@ -1947,5 +2035,262 @@ gtk_list_store_insert_with_valuesv (GtkListStore *list_store,
   gtk_tree_path_free (path);
 }
 
+/* GtkBuildable custom tag implementation
+ *
+ * <columns>
+ *   <column type="..."/>
+ *   <column type="..."/>
+ * </columns>
+ */
+typedef struct {
+  GtkBuilder *builder;
+  GObject *object;
+  GSList *column_type_names;
+  GType *column_types;
+  GValue *values;
+  gint *columns;
+  gint last_row;
+  gint n_columns;
+  gint row_column;
+  GQuark error_quark;
+  gboolean is_data;
+} SubParserData;
+
+static void
+list_store_start_element (GMarkupParseContext *context,
+			  const gchar         *element_name,
+			  const gchar        **names,
+			  const gchar        **values,
+			  gpointer             user_data,
+			  GError             **error)
+{
+  guint i;
+  SubParserData *data = (SubParserData*)user_data;
+
+  if (strcmp (element_name, "col") == 0)
+    {
+      int i, id = -1;
+
+      if (data->row_column >= data->n_columns)
+	g_set_error (error, data->error_quark, 0,
+		     "Too many columns, maximum is %d\n", data->n_columns - 1);
+
+      for (i = 0; names[i]; i++)
+	if (strcmp (names[i], "id") == 0)
+	  {
+	    errno = 0;
+	    id = atoi (values[i]);
+	    if (errno)
+	      g_set_error (error, data->error_quark, 0,
+			   "the id tag %s could not be converted to an integer", values[i]);
+	  }
+
+      if (id == -1)
+	g_set_error (error, data->error_quark, 0,
+		     "<col> needs an id attribute");
+
+      data->columns[data->row_column] = id;
+      data->row_column++;
+      data->is_data = TRUE;
+    }
+  else if (strcmp (element_name, "row") == 0)
+    ;
+  else if (strcmp (element_name, "column") == 0)
+    for (i = 0; names[i]; i++)
+      if (strcmp (names[i], "type") == 0)
+	data->column_type_names = g_slist_prepend (data->column_type_names,
+						   g_strdup (values[i]));
+  else if (strcmp (element_name, "columns") == 0)
+    ;
+  else if (strcmp (element_name, "data") == 0)
+    ;
+  else
+    g_set_error (error, data->error_quark, 0,
+		 "Unknown start tag: %s", element_name);
+}
+
+static void
+list_store_end_element (GMarkupParseContext *context,
+			const gchar         *element_name,
+			gpointer             user_data,
+			GError             **error)
+{
+  SubParserData *data = (SubParserData*)user_data;
+
+  g_assert (data->builder);
+  
+  if (strcmp (element_name, "row") == 0)
+    {
+      GtkTreeIter iter;
+      int i;
+
+      gtk_list_store_insert_with_valuesv (GTK_LIST_STORE (data->object),
+					  &iter,
+					  data->last_row,
+					  data->columns,
+					  data->values,
+					  data->row_column);
+      for (i = 0; i < data->row_column; i++)
+	g_value_unset (&data->values[i]);
+      g_free (data->values);
+      data->values = g_new0 (GValue, data->n_columns);
+      data->last_row++;
+      data->row_column = 0;
+    }
+  else if (strcmp (element_name, "columns") == 0)
+    {
+      GType *column_types;
+      GSList *l;
+      int i;
+      GType type;
+
+      data->column_type_names = g_slist_reverse (data->column_type_names);
+      column_types = g_new0 (GType, g_slist_length (data->column_type_names));
+
+      for (l = data->column_type_names, i = 0; l; l = l->next, i++)
+	{
+	  type = gtk_builder_get_type_from_name (data->builder, l->data);
+	  if (type == G_TYPE_INVALID)
+	    {
+	      g_warning ("Unknown type %s specified in treemodel %s",
+			 (const gchar*)l->data,
+			 gtk_buildable_get_name (GTK_BUILDABLE (data->object)));
+	      continue;
+	    }
+	  column_types[i] = type;
+
+	  g_free (l->data);
+	}
+
+      gtk_list_store_set_column_types (GTK_LIST_STORE (data->object), i,
+				       column_types);
+
+      g_free (column_types);
+    }
+  else if (strcmp (element_name, "col") == 0)
+    data->is_data = FALSE;
+  else if (strcmp (element_name, "data") == 0)
+    ;
+  else if (strcmp (element_name, "column") == 0)
+    ;
+  else
+    g_set_error (error, data->error_quark, 0,
+		 "Unknown end tag: %s", element_name);
+}
+
+static void
+list_store_text (GMarkupParseContext *context,
+		 const gchar         *text,
+		 gsize                text_len,
+		 gpointer             user_data,
+		 GError             **error)
+{
+  SubParserData *data = (SubParserData*)user_data;
+  gint i;
+  GError *tmp_error = NULL;
+  
+  if (!data->is_data)
+    return;
+
+  i = data->row_column - 1;
+
+  if (!gtk_builder_value_from_string_type (data->builder,
+					   data->column_types[i],
+					   text,
+					   &data->values[i],
+					   &tmp_error))
+    {
+      g_set_error (error,
+		   tmp_error->domain,
+		   tmp_error->code,
+		   "Could not convert '%s' to type %s: %s\n",
+		   text, g_type_name (data->column_types[i]),
+		   tmp_error->message);
+      g_error_free (tmp_error);
+    }
+}
+
+static const GMarkupParser list_store_parser =
+  {
+    list_store_start_element,
+    list_store_end_element,
+    list_store_text
+  };
+
+static gboolean
+gtk_list_store_buildable_custom_tag_start (GtkBuildable  *buildable,
+					   GtkBuilder    *builder,
+					   GObject       *child,
+					   const gchar   *tagname,
+					   GMarkupParser *parser,
+					   gpointer      *data)
+{
+  SubParserData *parser_data;
+
+  if (child)
+    return FALSE;
+
+  if (strcmp (tagname, "columns") == 0)
+    {
+
+      parser_data = g_slice_new0 (SubParserData);
+      parser_data->builder = builder;
+      parser_data->object = G_OBJECT (buildable);
+      parser_data->column_type_names = NULL;
+
+      *parser = list_store_parser;
+      *data = parser_data;
+      return TRUE;
+    }
+  else if (strcmp (tagname, "data") == 0)
+    {
+      gint n_columns = gtk_list_store_get_n_columns (GTK_TREE_MODEL (buildable));
+      if (n_columns == 0)
+	g_error ("Cannot append data to an empty model");
+
+      parser_data = g_slice_new0 (SubParserData);
+      parser_data->builder = builder;
+      parser_data->object = G_OBJECT (buildable);
+      parser_data->values = g_new0 (GValue, n_columns);
+      parser_data->columns = g_new0 (gint, n_columns);
+      parser_data->column_types = GTK_LIST_STORE (buildable)->column_headers;
+      parser_data->n_columns = n_columns;
+      parser_data->last_row = 0;
+      parser_data->error_quark = g_quark_from_static_string ("GtkListStore");
+      
+      *parser = list_store_parser;
+      *data = parser_data;
+      return TRUE;
+    }
+  else
+    g_warning ("Unknown custom list store tag: %s", tagname);
+  
+  return FALSE;
+}
+
+static void
+gtk_list_store_buildable_custom_tag_end (GtkBuildable *buildable,
+					 GtkBuilder   *builder,
+					 GObject      *child,
+					 const gchar  *tagname,
+					 gpointer     *data)
+{
+  SubParserData *sub = (SubParserData*)data;
+  
+  if (strcmp (tagname, "columns") == 0)
+    {
+      g_slist_free (sub->column_type_names);
+      g_slice_free (SubParserData, sub);
+    }
+  else if (strcmp (tagname, "data") == 0)
+    {
+      g_free (sub->columns);
+      g_free (sub->values);
+      g_slice_free (SubParserData, sub);
+    }
+  else
+    g_warning ("Unknown custom list store tag: %s", tagname);
+}
+
 #define __GTK_LIST_STORE_C__
 #include "gtkaliasdef.c"
