@@ -63,7 +63,7 @@ enum
 
 static gint signals[LAST_SIGNAL];
 
-G_DEFINE_TYPE (GtkMenuToolButton, gtk_menu_tool_button, GTK_TYPE_TOOL_BUTTON);
+G_DEFINE_TYPE (GtkMenuToolButton, gtk_menu_tool_button, GTK_TYPE_TOOL_BUTTON)
 
 static gboolean
 gtk_menu_tool_button_set_tooltip (GtkToolItem *tool_item,
@@ -596,6 +596,9 @@ gtk_menu_tool_button_get_menu (GtkMenuToolButton *button)
  * a tooltip on the whole #GtkMenuToolButton.
  *
  * Since: 2.6
+ *
+ * Deprecated: 2.12: Use gtk_menu_tool_button_set_arrow_tooltip_text()
+ * instead.
  **/
 void
 gtk_menu_tool_button_set_arrow_tooltip (GtkMenuToolButton *button,
@@ -608,5 +611,45 @@ gtk_menu_tool_button_set_arrow_tooltip (GtkMenuToolButton *button,
   gtk_tooltips_set_tip (tooltips, button->priv->arrow_button, tip_text, tip_private);
 }
 
+/**
+ * gtk_menu_tool_button_set_arrow_tooltip_text:
+ * @button: a #GtkMenuToolButton
+ * @text: text to be used as tooltip text for button's arrow button
+ *
+ * Sets the tooltip text to be used as tooltip for the arrow button which
+ * pops up the menu.  See gtk_tool_item_set_tooltip() for setting a tooltip
+ * on the whole #GtkMenuToolButton.
+ *
+ * Since: 2.12
+ **/
+void
+gtk_menu_tool_button_set_arrow_tooltip_text (GtkMenuToolButton *button,
+					     const gchar       *text)
+{
+  g_return_if_fail (GTK_IS_MENU_TOOL_BUTTON (button));
+
+  gtk_widget_set_tooltip_text (button->priv->arrow_button, text);
+}
+
+/**
+ * gtk_menu_tool_button_set_arrow_tooltip_markup:
+ * @button: a #GtkMenuToolButton
+ * @markup: markup text to be used as tooltip text for button's arrow button
+ *
+ * Sets the tooltip markup text to be used as tooltip for the arrow button
+ * which pops up the menu.  See gtk_tool_item_set_tooltip() for setting a
+ * tooltip on the whole #GtkMenuToolButton.
+ *
+ * Since: 2.12
+ **/
+void
+gtk_menu_tool_button_set_arrow_tooltip_markup (GtkMenuToolButton *button,
+					       const gchar       *markup)
+{
+  g_return_if_fail (GTK_IS_MENU_TOOL_BUTTON (button));
+
+  gtk_widget_set_tooltip_markup (button->priv->arrow_button, markup);
+}
+
 #define __GTK_MENU_TOOL_BUTTON_C__
 #include "gtkaliasdef.c"
