@@ -24,23 +24,23 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#ifndef GTK_DISABLE_DEPRECATED
+
 #ifndef __GTK_TOOLTIPS_H__
 #define __GTK_TOOLTIPS_H__
 
 #include <gtk/gtkwidget.h>
+#include <gtk/gtkwindow.h>
 
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
-
+G_BEGIN_DECLS
 
 #define GTK_TYPE_TOOLTIPS                  (gtk_tooltips_get_type ())
-#define GTK_TOOLTIPS(obj)                  (GTK_CHECK_CAST ((obj), GTK_TYPE_TOOLTIPS, GtkTooltips))
-#define GTK_TOOLTIPS_CLASS(klass)          (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_TOOLTIPS, GtkTooltipsClass))
-#define GTK_IS_TOOLTIPS(obj)               (GTK_CHECK_TYPE ((obj), GTK_TYPE_TOOLTIPS))
-#define GTK_IS_TOOLTIPS_CLASS(klass)       (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TOOLTIPS))
-#define GTK_TOOLTIPS_GET_CLASS(obj)        (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_TOOLTIPS, GtkTooltipsClass))
+#define GTK_TOOLTIPS(obj)                  (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_TOOLTIPS, GtkTooltips))
+#define GTK_TOOLTIPS_CLASS(klass)          (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_TOOLTIPS, GtkTooltipsClass))
+#define GTK_IS_TOOLTIPS(obj)               (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_TOOLTIPS))
+#define GTK_IS_TOOLTIPS_CLASS(klass)       (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TOOLTIPS))
+#define GTK_TOOLTIPS_GET_CLASS(obj)        (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_TOOLTIPS, GtkTooltipsClass))
 
 
 typedef struct _GtkTooltips	 GtkTooltips;
@@ -59,33 +59,38 @@ struct _GtkTooltips
 {
   GtkObject parent_instance;
 
+  /*< private >*/
   GtkWidget *tip_window;
   GtkWidget *tip_label;
   GtkTooltipsData *active_tips_data;
-  GList *tips_data_list;
+  GList *_tips_data_list; /* unused */
 
   guint   delay : 30;
   guint	  enabled : 1;
   guint   have_grab : 1;
+  guint   use_sticky_delay : 1;
   gint	  timer_tag;
-  gboolean use_sticky_delay;
   GTimeVal last_popdown;
 };
 
 struct _GtkTooltipsClass
 {
   GtkObjectClass parent_class;
+
+  /* Padding for future expansion */
+  void (*_gtk_reserved1) (void);
+  void (*_gtk_reserved2) (void);
+  void (*_gtk_reserved3) (void);
+  void (*_gtk_reserved4) (void);
 };
 
-GtkType		 gtk_tooltips_get_type	   (void) G_GNUC_CONST;
+GType		 gtk_tooltips_get_type	   (void) G_GNUC_CONST;
 GtkTooltips*	 gtk_tooltips_new	   (void);
 
 void		 gtk_tooltips_enable	   (GtkTooltips   *tooltips);
 void		 gtk_tooltips_disable	   (GtkTooltips   *tooltips);
-#ifndef GTK_DISABLE_DEPRECATED
 void		 gtk_tooltips_set_delay	   (GtkTooltips   *tooltips,
 					    guint	   delay);
-#endif /* GTK_DISABLE_DEPRECATED */
 void		 gtk_tooltips_set_tip	   (GtkTooltips   *tooltips,
 					    GtkWidget	  *widget,
 					    const gchar   *tip_text,
@@ -93,12 +98,12 @@ void		 gtk_tooltips_set_tip	   (GtkTooltips   *tooltips,
 GtkTooltipsData* gtk_tooltips_data_get	   (GtkWidget	  *widget);
 void             gtk_tooltips_force_window (GtkTooltips   *tooltips);
 
+gboolean         gtk_tooltips_get_info_from_tip_window (GtkWindow    *tip_window,
+                                                        GtkTooltips **tooltips,
+                                                        GtkWidget   **current_widget);
 
-void             _gtk_tooltips_show_tip    (GtkWidget    *widget);
-
-#ifdef __cplusplus
-}
-#endif /* __cplusplus */
-
+G_END_DECLS
 
 #endif /* __GTK_TOOLTIPS_H__ */
+
+#endif /* GTK_DISABLE_DEPRECATED */
