@@ -1,7 +1,7 @@
 /* GTK - The GIMP Toolkit
  * gtkcupsutils.h: Statemachine implementation of POST and GET 
- * cup calls which can be used to create a non-blocking cups API
- * Copyright (C) 2003, Red Hat, Inc.
+ * cups calls which can be used to create a non-blocking cups API
+ * Copyright (C) 2006, 2007 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -51,6 +51,11 @@ struct _GtkCupsResult
 {
   gchar *error_msg;
   ipp_t *ipp_response;
+  GtkCupsErrorType error_type;
+
+  /* some error types like HTTP_ERROR have a status and a code */
+  int error_status;            
+  int error_code;
 
   guint is_error : 1;
   guint is_ipp_response : 1;
@@ -60,28 +65,37 @@ struct _GtkCupsResult
 #define _GTK_CUPS_MAX_ATTEMPTS 10 
 #define _GTK_CUPS_MAX_CHUNK_SIZE 8192
 
-GtkCupsRequestStateFunc post_states[] = {_connect,
-                                         _post_send,
-                                         _post_write_request,
-                                         _post_write_data,
-                                         _post_check,
-                                         _post_read_response};
+static GtkCupsRequestStateFunc post_states[] = {
+  _connect,
+  _post_send,
+  _post_write_request,
+  _post_write_data,
+  _post_check,
+  _post_read_response
+};
 
-GtkCupsRequestStateFunc get_states[] = {_connect,
-                                        _get_send,
-                                        _get_check,
-                                        _get_read_data};
+static GtkCupsRequestStateFunc get_states[] = {
+  _connect,
+  _get_send,
+  _get_check,
+  _get_read_data
+};
 
 static void
-gtk_cups_result_set_error (GtkCupsResult *result, 
-                           const char *error_msg,
+gtk_cups_result_set_error (GtkCupsResult    *result,
+                           GtkCupsErrorType  error_type,
+                           int               error_status,
+                           int               error_code, 
+                           const char       *error_msg,
 			   ...)
 {
   va_list args;
 
   result->is_ipp_response = FALSE;
-
   result->is_error = TRUE;
+  result->error_type = error_type;
+  result->error_status = error_status;
+  result->error_code = error_code;
 
   va_start (args, error_msg);
   result->error_msg = g_strdup_vprintf (error_msg, args);
@@ -89,12 +103,12 @@ gtk_cups_result_set_error (GtkCupsResult *result,
 }
 
 GtkCupsRequest *
-gtk_cups_request_new (http_t *connection,
-                      GtkCupsRequestType req_type, 
-                      gint operation_id,
-                      GIOChannel *data_io,
-                      const char *server,
-                      const char *resource)
+gtk_cups_request_new (http_t             *connection,
+                      GtkCupsRequestType  req_type, 
+                      gint                operation_id,
+                      GIOChannel         *data_io,
+                      const char         *server,
+                      const char         *resource)
 {
   GtkCupsRequest *request;
   cups_lang_t *language;
@@ -114,7 +128,7 @@ gtk_cups_request_new (http_t *connection,
    if (server)
     request->server = g_strdup (server);
   else
-    request->server = g_strdup (cupsServer());
+    request->server = g_strdup (cupsServer ());
 
 
   if (resource)
@@ -130,7 +144,9 @@ gtk_cups_request_new (http_t *connection,
   else
     {
       request->http = NULL;
-      request->http = httpConnectEncrypt (request->server, ippPort(), cupsEncryption());
+      request->http = httpConnectEncrypt (request->server, 
+                                          ippPort (), 
+                                          cupsEncryption ());
 
       if (request->http)
         httpBlocking (request->http, 0);
@@ -143,7 +159,7 @@ gtk_cups_request_new (http_t *connection,
   request->attempts = 0;
   request->data_io = data_io;
 
-  request->ipp_request = ippNew();
+  request->ipp_request = ippNew ();
   request->ipp_request->request.op.operation_id = operation_id;
   request->ipp_request->request.op.request_id = 1;
 
@@ -159,7 +175,7 @@ gtk_cups_request_new (http_t *connection,
 
   gtk_cups_request_ipp_add_string (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
                                    "requesting-user-name",
-                                   NULL, cupsUser());
+                                   NULL, cupsUser ());
   
   cupsLangFree (language);
 
@@ -181,8 +197,10 @@ void
 gtk_cups_request_free (GtkCupsRequest *request)
 {
   if (request->own_http)
-    if (request->http)
-      httpClose (request->http);
+    {
+      if (request->http)
+        httpClose (request->http);
+    }
   
   if (request->ipp_request)
     ippDelete (request->ipp_request);
@@ -199,14 +217,20 @@ gboolean
 gtk_cups_request_read_write (GtkCupsRequest *request)
 {
   if (request->type == GTK_CUPS_POST)
-    post_states[request->state](request);
+    post_states[request->state] (request);
   else if (request->type == GTK_CUPS_GET)
-    get_states[request->state](request);
+    get_states[request->state] (request);
 
   if (request->attempts > _GTK_CUPS_MAX_ATTEMPTS && 
       request->state != GTK_CUPS_REQUEST_DONE)
     {
-      gtk_cups_result_set_error (request->result, "Too many failed attempts");
+      /* TODO: should add a status or error code for too many failed attempts */
+      gtk_cups_result_set_error (request->result, 
+                                 GTK_CUPS_ERROR_GENERAL,
+                                 0,
+                                 0, 
+                                 "Too many failed attempts");
+
       request->state = GTK_CUPS_REQUEST_DONE;
       request->poll_state = GTK_CUPS_HTTP_IDLE;
     }
@@ -217,9 +241,7 @@ gtk_cups_request_read_write (GtkCupsRequest *request)
       return TRUE;
     }
   else
-    {
-      return FALSE;
-    }
+    return FALSE;
 }
 
 GtkCupsPollState 
@@ -238,11 +260,11 @@ gtk_cups_request_get_result (GtkCupsRequest *request)
 
 void            
 gtk_cups_request_ipp_add_string (GtkCupsRequest *request,
-                                 ipp_tag_t group,
-                                 ipp_tag_t tag,
-                                 const char *name,
-                                 const char *charset,
-                                 const char *value)
+                                 ipp_tag_t       group,
+                                 ipp_tag_t       tag,
+                                 const char     *name,
+                                 const char     *charset,
+                                 const char     *value)
 {
   ippAddString (request->ipp_request,
                 group,
@@ -253,13 +275,13 @@ gtk_cups_request_ipp_add_string (GtkCupsRequest *request,
 }
 
 void            
-gtk_cups_request_ipp_add_strings (GtkCupsRequest *request,
-				  ipp_tag_t group,
-				  ipp_tag_t tag,
-				  const char *name,
-				  int num_values,
-				  const char *charset,
-				  const char * const *values)
+gtk_cups_request_ipp_add_strings (GtkCupsRequest    *request,
+				  ipp_tag_t          group,
+				  ipp_tag_t          tag,
+				  const char        *name,
+				  int                num_values,
+				  const char        *charset,
+				  const char *const *values)
 {
   ippAddStrings (request->ipp_request,
 		 group,
@@ -278,41 +300,40 @@ typedef struct
   ipp_tag_t	value_tag;
 } ipp_option_t;
 
-static const ipp_option_t ipp_options[] =
-			{
-			  { "blackplot",		IPP_TAG_BOOLEAN },
-			  { "brightness",		IPP_TAG_INTEGER },
-			  { "columns",			IPP_TAG_INTEGER },
-			  { "copies",			IPP_TAG_INTEGER },
-			  { "finishings",		IPP_TAG_ENUM },
-			  { "fitplot",			IPP_TAG_BOOLEAN },
-			  { "gamma",			IPP_TAG_INTEGER },
-			  { "hue",			IPP_TAG_INTEGER },
-			  { "job-k-limit",		IPP_TAG_INTEGER },
-			  { "job-page-limit",		IPP_TAG_INTEGER },
-			  { "job-priority",		IPP_TAG_INTEGER },
-			  { "job-quota-period",		IPP_TAG_INTEGER },
-			  { "landscape",		IPP_TAG_BOOLEAN },
-			  { "media",			IPP_TAG_KEYWORD },
-			  { "mirror",			IPP_TAG_BOOLEAN },
-			  { "natural-scaling",		IPP_TAG_INTEGER },
-			  { "number-up",		IPP_TAG_INTEGER },
-			  { "orientation-requested",	IPP_TAG_ENUM },
-			  { "page-bottom",		IPP_TAG_INTEGER },
-			  { "page-left",		IPP_TAG_INTEGER },
-			  { "page-ranges",		IPP_TAG_RANGE },
-			  { "page-right",		IPP_TAG_INTEGER },
-			  { "page-top",			IPP_TAG_INTEGER },
-			  { "penwidth",			IPP_TAG_INTEGER },
-			  { "ppi",			IPP_TAG_INTEGER },
-			  { "prettyprint",		IPP_TAG_BOOLEAN },
-			  { "printer-resolution",	IPP_TAG_RESOLUTION },
-			  { "print-quality",		IPP_TAG_ENUM },
-			  { "saturation",		IPP_TAG_INTEGER },
-			  { "scaling",			IPP_TAG_INTEGER },
-			  { "sides",			IPP_TAG_KEYWORD },
-			  { "wrap",			IPP_TAG_BOOLEAN }
-			};
+static const ipp_option_t ipp_options[] = {
+  { "blackplot",		IPP_TAG_BOOLEAN },
+  { "brightness",		IPP_TAG_INTEGER },
+  { "columns",			IPP_TAG_INTEGER },
+  { "copies",			IPP_TAG_INTEGER },
+  { "finishings",		IPP_TAG_ENUM },
+  { "fitplot",			IPP_TAG_BOOLEAN },
+  { "gamma",			IPP_TAG_INTEGER },
+  { "hue",			IPP_TAG_INTEGER },
+  { "job-k-limit",		IPP_TAG_INTEGER },
+  { "job-page-limit",		IPP_TAG_INTEGER },
+  { "job-priority",		IPP_TAG_INTEGER },
+  { "job-quota-period",		IPP_TAG_INTEGER },
+  { "landscape",		IPP_TAG_BOOLEAN },
+  { "media",			IPP_TAG_KEYWORD },
+  { "mirror",			IPP_TAG_BOOLEAN },
+  { "natural-scaling",		IPP_TAG_INTEGER },
+  { "number-up",		IPP_TAG_INTEGER },
+  { "orientation-requested",	IPP_TAG_ENUM },
+  { "page-bottom",		IPP_TAG_INTEGER },
+  { "page-left",		IPP_TAG_INTEGER },
+  { "page-ranges",		IPP_TAG_RANGE },
+  { "page-right",		IPP_TAG_INTEGER },
+  { "page-top",			IPP_TAG_INTEGER },
+  { "penwidth",			IPP_TAG_INTEGER },
+  { "ppi",			IPP_TAG_INTEGER },
+  { "prettyprint",		IPP_TAG_BOOLEAN },
+  { "printer-resolution",	IPP_TAG_RESOLUTION },
+  { "print-quality",		IPP_TAG_ENUM },
+  { "saturation",		IPP_TAG_INTEGER },
+  { "scaling",			IPP_TAG_INTEGER },
+  { "sides",			IPP_TAG_KEYWORD },
+  { "wrap",			IPP_TAG_BOOLEAN }
+};
 
 
 static ipp_tag_t
@@ -325,14 +346,14 @@ _find_option_tag (const gchar *option)
   result = IPP_TAG_ZERO;
 
   lower_bound = 0;
-  upper_bound = num_options = (int)(sizeof(ipp_options) / sizeof(ipp_options[0])) - 1;
+  upper_bound = num_options = (int) G_N_ELEMENTS (ipp_options) - 1;
   
   while (1)
     {
       int match;
       current_option = (int) (((upper_bound - lower_bound) / 2) + lower_bound);
 
-      match = strcasecmp(option, ipp_options[current_option].name);
+      match = strcasecmp (option, ipp_options[current_option].name);
       if (match == 0)
         {
 	  result = ipp_options[current_option].value_tag;
@@ -361,15 +382,20 @@ _find_option_tag (const gchar *option)
     }
 }
 
+/*
+ * Note that this function uses IPP_TAG_JOB, so it is
+ * only suitable for IPP Group 2 attributes.
+ * See RFC 2911.
+ */
 void
 gtk_cups_request_encode_option (GtkCupsRequest *request,
-                                const gchar *option,
-				const gchar *value)
+                                const gchar    *option,
+			        const gchar    *value)
 {
   ipp_tag_t option_tag;
 
-  g_assert (option != NULL);
-  g_assert (value != NULL);
+  g_return_if_fail (option != NULL);
+  g_return_if_fail (value != NULL);
 
   option_tag = _find_option_tag (option);
 
@@ -388,7 +414,7 @@ gtk_cups_request_encode_option (GtkCupsRequest *request,
       case IPP_TAG_INTEGER:
       case IPP_TAG_ENUM:
         ippAddInteger (request->ipp_request,
-                       IPP_TAG_OPERATION,
+                       IPP_TAG_JOB,
                        option_tag,
                        option,
                        strtol (value, NULL, 0));
@@ -397,16 +423,18 @@ gtk_cups_request_encode_option (GtkCupsRequest *request,
       case IPP_TAG_BOOLEAN:
         {
           char b;
-          b = 0;
-          if (!strcasecmp(value, "true") ||
-	      !strcasecmp(value, "on") ||
-	      !strcasecmp(value, "yes")) 
+          
+          if (strcasecmp (value, "true") == 0 ||
+	      strcasecmp (value, "on") == 0 ||
+	      strcasecmp (value, "yes") == 0) 
 	    b = 1;
-	  
-          ippAddBoolean(request->ipp_request,
-                        IPP_TAG_OPERATION,
-                        option,
-                        b);
+	  else
+            b = 0;
+
+          ippAddBoolean (request->ipp_request,
+                         IPP_TAG_JOB,
+                         option,
+                         b);
         
           break;
         }
@@ -423,12 +451,12 @@ gtk_cups_request_encode_option (GtkCupsRequest *request,
 	      s = (char *)value;
 	    }
 	  else
-	    lower = strtol(value, &s, 0);
+	    lower = strtol (value, &s, 0);
 
 	  if (*s == '-')
 	    {
 	      if (s[1])
-		upper = strtol(s + 1, NULL, 0);
+		upper = strtol (s + 1, NULL, 0);
 	      else
 		upper = 2147483647;
             }
@@ -436,7 +464,7 @@ gtk_cups_request_encode_option (GtkCupsRequest *request,
 	    upper = lower;
          
           ippAddRange (request->ipp_request,
-                       IPP_TAG_OPERATION,
+                       IPP_TAG_JOB,
                        option,
                        lower,
                        upper);
@@ -451,20 +479,20 @@ gtk_cups_request_encode_option (GtkCupsRequest *request,
           int yres;
           ipp_res_t units;
           
-          xres = strtol(value, &s, 0);
+          xres = strtol (value, &s, 0);
 
 	  if (*s == 'x')
-	    yres = strtol(s + 1, &s, 0);
+	    yres = strtol (s + 1, &s, 0);
 	  else
 	    yres = xres;
 
-	  if (strcasecmp(s, "dpc") == 0)
+	  if (strcasecmp (s, "dpc") == 0)
             units = IPP_RES_PER_CM;
           else
             units = IPP_RES_PER_INCH;
           
           ippAddResolution (request->ipp_request,
-                            IPP_TAG_OPERATION,
+                            IPP_TAG_JOB,
                             option,
                             units,
                             xres,
@@ -474,12 +502,80 @@ gtk_cups_request_encode_option (GtkCupsRequest *request,
         }
 
       default:
-        ippAddString (request->ipp_request,
-                      IPP_TAG_OPERATION,
-                      option_tag,
-                      option,
-                      NULL,
-                      value);
+        {
+          char *values;
+          char *s;
+          int in_quotes;
+          char *next;
+          GPtrArray *strings;
+          
+          values = g_strdup (value);
+          strings = NULL;
+	  in_quotes = 0;
+
+          for (s = values, next = s; *s != '\0'; s++)
+            {
+              if (in_quotes != 2 && *s == '\'')
+                {
+                  /* skip quoted value */
+                  if (in_quotes == 0)
+                    in_quotes = 1;
+                  else
+                    in_quotes = 0;
+                }
+              else if (in_quotes != 1 && *s == '\"')
+                {
+                  /* skip quoted value */
+                  if (in_quotes == 0)
+                    in_quotes = 2;
+                  else
+                    in_quotes = 0;
+                }
+              else if (in_quotes == 0 && *s == ',')
+                {
+                  /* found delimiter, add to value array */
+                  *s = '\0';
+                  if (strings == NULL)
+                    strings = g_ptr_array_new ();
+                  g_ptr_array_add (strings, next);
+                  next = s + 1;
+                }
+              else if (in_quotes == 0 && *s == '\\' && s[1] != '\0')
+                {
+                  /* skip escaped character */
+                  s++;
+                }
+            }
+          
+          if (strings == NULL)
+            {
+              /* single value */
+              ippAddString (request->ipp_request,
+                            IPP_TAG_JOB,
+                            option_tag,
+                            option,
+                            NULL,
+                            value);
+            }
+          else
+            {
+              /* multiple values */
+              
+              /* add last value */
+              g_ptr_array_add (strings, next);
+              
+              ippAddStrings (request->ipp_request,
+                             IPP_TAG_JOB,
+                             option_tag,
+                             option,
+                             strings->len,
+                             NULL,
+                             (const char **) strings->pdata);
+              g_ptr_array_free (strings, TRUE);
+            }
+
+          g_free (values);
+        }
 
         break;
     }
@@ -493,7 +589,9 @@ _connect (GtkCupsRequest *request)
 
   if (request->http == NULL)
     {
-      request->http = httpConnectEncrypt (request->server, ippPort(), cupsEncryption());
+      request->http = httpConnectEncrypt (request->server, 
+                                          ippPort (), 
+                                          cupsEncryption ());
 
       if (request->http == NULL)
         request->attempts++;
@@ -528,28 +626,31 @@ _post_send (GtkCupsRequest *request)
   if (request->data_io != NULL)
     {
       fstat (g_io_channel_unix_get_fd (request->data_io), &data_info);
-      sprintf (length, "%lu", (unsigned long)ippLength(request->ipp_request) + data_info.st_size);
+      sprintf (length, "%lu", (unsigned long) ippLength (request->ipp_request) + data_info.st_size);
     }
   else
-    {
-      sprintf (length, "%lu", (unsigned long)ippLength(request->ipp_request));
-    }
+    sprintf (length, "%lu", (unsigned long) ippLength (request->ipp_request));
 	
-  httpClearFields(request->http);
-  httpSetField(request->http, HTTP_FIELD_CONTENT_LENGTH, length);
-  httpSetField(request->http, HTTP_FIELD_CONTENT_TYPE, "application/ipp");
+  httpClearFields (request->http);
+  httpSetField (request->http, HTTP_FIELD_CONTENT_LENGTH, length);
+  httpSetField (request->http, HTTP_FIELD_CONTENT_TYPE, "application/ipp");
 #ifdef HAVE_HTTP_AUTHSTRING
-  httpSetField(request->http, HTTP_FIELD_AUTHORIZATION, request->http->authstring);
+  httpSetField (request->http, HTTP_FIELD_AUTHORIZATION, request->http->authstring);
 #endif
 
-  if (httpPost(request->http, request->resource))
+  if (httpPost (request->http, request->resource))
     {
-      if (httpReconnect(request->http))
+      if (httpReconnect (request->http))
         {
           request->state = GTK_CUPS_POST_DONE;
           request->poll_state = GTK_CUPS_HTTP_IDLE;
 
-          gtk_cups_result_set_error (request->result, "Failed Post");
+          /* TODO: should add a status or error code for failed post */
+          gtk_cups_result_set_error (request->result,
+                                     GTK_CUPS_ERROR_GENERAL,
+                                     0,
+                                     0,
+                                     "Failed Post");
         }
 
       request->attempts++;
@@ -572,14 +673,20 @@ _post_write_request (GtkCupsRequest *request)
 
   request->poll_state = GTK_CUPS_HTTP_WRITE;
   
-  ipp_status = ippWrite(request->http, request->ipp_request);
+  ipp_status = ippWrite (request->http, request->ipp_request);
 
   if (ipp_status == IPP_ERROR)
     {
+      int cups_error = cupsLastError ();
       request->state = GTK_CUPS_POST_DONE;
       request->poll_state = GTK_CUPS_HTTP_IDLE;
  
-      gtk_cups_result_set_error (request->result, "%s",ippErrorString (cupsLastError ()));
+      gtk_cups_result_set_error (request->result, 
+                                 GTK_CUPS_ERROR_IPP,
+                                 ipp_status,
+                                 cups_error,
+                                 "%s", 
+                                 ippErrorString (cups_error));
       return;
     }
 
@@ -608,7 +715,7 @@ _post_write_data (GtkCupsRequest *request)
   request->poll_state = GTK_CUPS_HTTP_WRITE;
   
   if (httpCheck (request->http))
-    http_status = httpUpdate(request->http);
+    http_status = httpUpdate (request->http);
   else
     http_status = request->last_status;
 
@@ -635,7 +742,12 @@ _post_write_data (GtkCupsRequest *request)
           request->state = GTK_CUPS_POST_DONE;
 	  request->poll_state = GTK_CUPS_HTTP_IDLE;
      
-          gtk_cups_result_set_error (request->result, "Error reading from cache file: %s", error->message);
+          gtk_cups_result_set_error (request->result,
+                                     GTK_CUPS_ERROR_IO,
+                                     io_status,
+                                     error->code, 
+                                     "Error reading from cache file: %s",
+                                     error->message);
 
 	  g_error_free (error);
           return;
@@ -651,15 +763,24 @@ _post_write_data (GtkCupsRequest *request)
 
 
 #if HAVE_CUPS_API_1_2
-      if (httpWrite2(request->http, buffer, bytes) < bytes)
+      if (httpWrite2 (request->http, buffer, bytes) < bytes)
 #else
-      if (httpWrite(request->http, buffer, (int) bytes) < bytes)
+      if (httpWrite (request->http, buffer, (int) bytes) < bytes)
 #endif /* HAVE_CUPS_API_1_2 */
         {
+          int http_errno;
+
+          http_errno = httpError (request->http);
+
           request->state = GTK_CUPS_POST_DONE;
 	  request->poll_state = GTK_CUPS_HTTP_IDLE;
      
-          gtk_cups_result_set_error (request->result, "Error writting to socket in Post %s", strerror (httpError (request->http)));
+          gtk_cups_result_set_error (request->result,
+                                     GTK_CUPS_ERROR_HTTP,
+                                     http_status,
+                                     http_errno, 
+                                     "Error writing to socket in Post %s", 
+                                     g_strerror (http_errno));
           return;
         }
     }
@@ -692,17 +813,21 @@ _post_check (GtkCupsRequest *request)
       request->state = GTK_CUPS_POST_DONE;
       request->poll_state = GTK_CUPS_HTTP_IDLE;
       
-      gtk_cups_result_set_error (request->result, "Can't prompt for authorization");
+      /* TODO: create a not implemented error code */
+      gtk_cups_result_set_error (request->result, 
+                                 GTK_CUPS_ERROR_GENERAL,
+                                 0,
+                                 0,
+                                 "Can't prompt for authorization");
       return;
     }
   else if (http_status == HTTP_ERROR)
     {
+      int error = httpError (request->http);
 #ifdef G_OS_WIN32
-      if (request->http->error != WSAENETDOWN && 
-          request->http->error != WSAENETUNREACH)
+      if (error != WSAENETDOWN && error != WSAENETUNREACH)
 #else
-      if (request->http->error != ENETDOWN && 
-          request->http->error != ENETUNREACH)
+      if (error != ENETDOWN && error != ENETUNREACH)	  
 #endif /* G_OS_WIN32 */
         {
           request->attempts++;
@@ -713,27 +838,32 @@ _post_check (GtkCupsRequest *request)
           request->state = GTK_CUPS_POST_DONE;
           request->poll_state = GTK_CUPS_HTTP_IDLE;
      
-          gtk_cups_result_set_error (request->result, "Unknown HTTP error");
+          gtk_cups_result_set_error (request->result,
+                                     GTK_CUPS_ERROR_HTTP,
+                                     http_status,
+                                     error, 
+                                     "Unknown HTTP error");
+
           return;
         }
     }
-/* TODO: detect ssl in configure.ac */
-#if HAVE_SSL
   else if (http_status == HTTP_UPGRADE_REQUIRED)
     {
       /* Flush any error message... */
       httpFlush (request->http);
 
+      cupsSetEncryption (HTTP_ENCRYPT_REQUIRED);
+      request->state = GTK_CUPS_POST_CONNECT;
+
       /* Reconnect... */
       httpReconnect (request->http);
 
       /* Upgrade with encryption... */
-      httpEncryption(request->http, HTTP_ENCRYPT_REQUIRED);
+      httpEncryption (request->http, HTTP_ENCRYPT_REQUIRED);
  
       request->attempts++;
       goto again;
     }
-#endif 
   else if (http_status != HTTP_OK)
     {
       int http_errno;
@@ -745,16 +875,21 @@ _post_check (GtkCupsRequest *request)
       else
         {
           request->state = GTK_CUPS_POST_DONE;
-          gtk_cups_result_set_error (request->result, "HTTP Error in POST %s", strerror (http_errno));
+          gtk_cups_result_set_error (request->result,
+                                     GTK_CUPS_ERROR_HTTP,
+                                     http_status,
+                                     http_errno, 
+                                     "HTTP Error in POST %s", 
+                                     g_strerror (http_errno));
          request->poll_state = GTK_CUPS_HTTP_IDLE;
  
-          httpFlush(request->http); 
+          httpFlush (request->http); 
           return;
         }
 
       request->poll_state = GTK_CUPS_HTTP_IDLE;
        
-      httpFlush(request->http); 
+      httpFlush (request->http); 
       
       request->last_status = HTTP_CONTINUE;
       httpClose (request->http);
@@ -794,7 +929,13 @@ _post_read_response (GtkCupsRequest *request)
 
   if (ipp_status == IPP_ERROR)
     {
-      gtk_cups_result_set_error (request->result, "%s", ippErrorString (cupsLastError()));
+      int ipp_error = cupsLastError ();
+      gtk_cups_result_set_error (request->result,  
+                                 GTK_CUPS_ERROR_IPP,
+                                 ipp_status,
+                                 ipp_error,
+                                 "%s",
+                                 ippErrorString (ipp_error));
       
       ippDelete (request->result->ipp_response);
       request->result->ipp_response = NULL;
@@ -819,26 +960,36 @@ _get_send (GtkCupsRequest *request)
 
   if (request->data_io == NULL)
     {
-      gtk_cups_result_set_error (request->result, "Get requires an open io channel");
+      gtk_cups_result_set_error (request->result,
+                                 GTK_CUPS_ERROR_IO,
+                                 G_IO_STATUS_ERROR,
+                                 G_IO_CHANNEL_ERROR_FAILED, 
+                                 "Get requires an open io channel");
+
       request->state = GTK_CUPS_GET_DONE;
       request->poll_state = GTK_CUPS_HTTP_IDLE;
 
       return;
     }
 
-  httpClearFields(request->http);
+  httpClearFields (request->http);
 #ifdef HAVE_HTTP_AUTHSTRING
-  httpSetField(request->http, HTTP_FIELD_AUTHORIZATION, request->http->authstring);
+  httpSetField (request->http, HTTP_FIELD_AUTHORIZATION, request->http->authstring);
 #endif
 
-  if (httpGet(request->http, request->resource))
+  if (httpGet (request->http, request->resource))
     {
-      if (httpReconnect(request->http))
+      if (httpReconnect (request->http))
         {
           request->state = GTK_CUPS_GET_DONE;
           request->poll_state = GTK_CUPS_HTTP_IDLE;
-	  
-          gtk_cups_result_set_error (request->result, "Failed Get");
+	 
+          /* TODO: should add a status or error code for failed GET */ 
+          gtk_cups_result_set_error (request->result, 
+                                     GTK_CUPS_ERROR_GENERAL,
+                                     0,
+                                     0,
+                                     "Failed Get");
         }
 
       request->attempts++;
@@ -875,27 +1026,32 @@ _get_check (GtkCupsRequest *request)
       g_warning ("NOT IMPLEMENTED: We need to prompt for authorization in a non blocking manner");
       request->state = GTK_CUPS_GET_DONE;
       request->poll_state = GTK_CUPS_HTTP_IDLE;
- 
-      gtk_cups_result_set_error (request->result, "Can't prompt for authorization");
+
+      /* TODO: should add a status or error code for not implemented */ 
+      gtk_cups_result_set_error (request->result, 
+                                 GTK_CUPS_ERROR_GENERAL,
+                                 0,
+                                 0,
+                                 "Can't prompt for authorization");
       return;
     }
-/* TODO: detect ssl in configure.ac */
-#if HAVE_SSL
   else if (http_status == HTTP_UPGRADE_REQUIRED)
     {
       /* Flush any error message... */
       httpFlush (request->http);
 
+      cupsSetEncryption (HTTP_ENCRYPT_REQUIRED);
+      request->state = GTK_CUPS_POST_CONNECT;
+
       /* Reconnect... */
       httpReconnect (request->http);
 
       /* Upgrade with encryption... */
-      httpEncryption(request->http, HTTP_ENCRYPT_REQUIRED);
+      httpEncryption (request->http, HTTP_ENCRYPT_REQUIRED);
  
       request->attempts++;
       goto again;
     }
-#endif 
   else if (http_status != HTTP_OK)
     {
       int http_errno;
@@ -907,9 +1063,14 @@ _get_check (GtkCupsRequest *request)
       else
         {
           request->state = GTK_CUPS_GET_DONE;
-          gtk_cups_result_set_error (request->result, "HTTP Error in GET %s", strerror (http_errno));
+          gtk_cups_result_set_error (request->result,
+                                     GTK_CUPS_ERROR_HTTP,
+                                     http_status,
+                                     http_errno, 
+                                     "HTTP Error in GET %s", 
+                                     g_strerror (http_errno));
           request->poll_state = GTK_CUPS_HTTP_IDLE;
-          httpFlush(request->http);
+          httpFlush (request->http);
 
           return;
         }
@@ -955,9 +1116,9 @@ _get_read_data (GtkCupsRequest *request)
   request->poll_state = GTK_CUPS_HTTP_READ;
 
 #if HAVE_CUPS_API_1_2
-  bytes = httpRead2(request->http, buffer, sizeof(buffer));
+  bytes = httpRead2 (request->http, buffer, sizeof (buffer));
 #else
-  bytes = httpRead(request->http, buffer, sizeof(buffer));
+  bytes = httpRead (request->http, buffer, sizeof (buffer));
 #endif /* HAVE_CUPS_API_1_2 */
 
   GTK_NOTE (PRINTING,
@@ -983,7 +1144,11 @@ _get_read_data (GtkCupsRequest *request)
       request->state = GTK_CUPS_POST_DONE;
       request->poll_state = GTK_CUPS_HTTP_IDLE;
     
-      gtk_cups_result_set_error (request->result, error->message);
+      gtk_cups_result_set_error (request->result,
+                                 GTK_CUPS_ERROR_IO,
+                                 io_status,
+                                 error->code, 
+                                 error->message);
       g_error_free (error);
     }
 }
@@ -1006,6 +1171,24 @@ gtk_cups_result_get_response (GtkCupsResult *result)
   return result->ipp_response;
 }
 
+GtkCupsErrorType
+gtk_cups_result_get_error_type (GtkCupsResult *result)
+{
+  return result->error_type;
+}
+
+int
+gtk_cups_result_get_error_status (GtkCupsResult *result)
+{
+  return result->error_status;
+}
+
+int
+gtk_cups_result_get_error_code (GtkCupsResult *result)
+{
+  return result->error_code;
+}
+
 const char *
 gtk_cups_result_get_error_string (GtkCupsResult *result)
 {
